#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import logging
import os
import pathlib
from email.mime.text import MIMEText
from smtplib import SMTP

import deprecation
from importlib.metadata import version
import pwd

# retro compatibility, force import methods from actual perl_utils module
from ensembl.production.core.perl_utils import \
    dict_to_perl_string as pdict_to_perl_string, \
    perl_string_to_python as pperl_string_to_python, \
    escape_perl_string as pescape_perl_string, \
    list_to_perl_string as plist_to_perl_string

here = pathlib.Path(__file__).parents[4].resolve()

__version__ = version('ensembl-prodinf-core')

@deprecation.deprecated(deprecated_in="2.0.2", removed_in="3.0.0",
                        current_version=__version__,
                        details="Use the ensembl.production.core.perl_utils.perl_string_to_python instead")
def perl_string_to_python(s):
    return pperl_string_to_python(s)


@deprecation.deprecated(deprecated_in="2.0.2", removed_in="3.0.0",
                        current_version=__version__,
                        details="Use the ensembl.production.core.perl_utils.dict_to_perl_string instead")
def dict_to_perl_string(input_dict):
    return pdict_to_perl_string(input_dict)


@deprecation.deprecated(deprecated_in="2.0.2", removed_in="3.0.0",
                        current_version=__version__,
                        details="Use the ensembl.production.core.perl_utils.list_to_perl_string instead")
def list_to_perl_string(input_list):
    return plist_to_perl_string(input_list)


@deprecation.deprecated(deprecated_in="2.0.2", removed_in="3.0.0",
                        current_version=__version__,
                        details="Use the ensembl.production.core.perl_utils.escape_perl_string instead")
def escape_perl_string(s):
    return pescape_perl_string(s)


def get_default_user():
    """Method to obtain the current user. This can be complicated when running Docker containers"""
    default_user = None
    for name in ('LOGNAME', 'USER', 'LNAME', 'USERNAME'):
        user = os.environ.get(name)
        if user:
            default_user = user
            break
    if not default_user:
        default_user = pwd.getpwuid(os.getuid()).pw_name
    return default_user


def send_email(**kwargs):
    """ Utility method for sending an email"""
    logger = kwargs.get('logger', logging)
    from_address = kwargs.get('from_email_address', 'ensembl-production@ebi.ac.uk')
    msg = MIMEText(kwargs['body'])
    msg['Subject'] = kwargs['subject']
    msg['From'] = from_address
    msg['To'] = kwargs['to_address']
    smtp_server = kwargs.get('smtp_server', 'localhost')
    to_address = kwargs['to_address']
    logger.debug(
        'sendmail server: {} - Message from: {}, to: {}, subject: {}'.format(smtp_server, from_address, to_address,
                                                                             msg['Subject']))
    s = SMTP(smtp_server)
    s.sendmail(from_address, (to_address,), msg.as_string())
    s.quit()


def json_decode_error_context(error):
    beg = max(0, error.pos - 25)
    end = min(len(error.doc), error.pos + 25)
    return f"{error}. --> {error.doc[beg:end]} <--"



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# TODO clean up obsolete methods not available since we don't have mysql ssh access anymore
"""
Utilities for dealing with REST and database servers
"""
import re
from subprocess import Popen, PIPE, TimeoutExpired
from urllib.parse import urlparse


db_netloc_re = re.compile(r'^.+@.+:\d+$')
email_re = re.compile(r'^\w+([\.-]?\w+)*@\w+([\.-]?\w+)*(\.\w{2,3})+$')


def assert_http_uri(uri):
    """Check supplied URI matches http"""
    parsed_uri = urlparse(uri)
    if not (parsed_uri.scheme.startswith('http') and parsed_uri.netloc and parsed_uri.path):
        raise ValueError("Endpoint URL doesn't match pattern: http(s)://server-name(:port)/")

def assert_mysql_uri(uri):
    """Check supplied URI matches MySQL server"""
    parsed_uri = urlparse(uri)
    valid_netloc = db_netloc_re.match(parsed_uri.netloc)
    if not (parsed_uri.scheme == 'mysql' and valid_netloc and len(parsed_uri.path) == 1):
        raise ValueError("MySQL URL doesn't match pattern: mysql://user(:pass)@server:port/")

def assert_mysql_db_uri(uri):
    """Check supplied URI matches MySQL database"""
    parsed_uri = urlparse(uri)
    valid_netloc = db_netloc_re.match(parsed_uri.netloc)
    if not (parsed_uri.scheme == 'mysql' and valid_netloc and len(parsed_uri.path) > 2):
        raise ValueError("MySQL database URL doesn't match pattern: mysql://user(:pass)@server:port/prod_db_name")

def assert_email(email):
    """Check supplied string is an email address"""
    if not email_re.match(email):
        raise ValueError("Email doesn't match pattern: user@domain")

def get_load(host=None):
    """Find load by on the supplied host by ssh (or on localhost if no host is supplied)
    """
    # load obtained from uptime
    status = run_process('uptime', process_uptime, host)
    return status

def get_file_sizes(host=None, dir_name=None):
    """Find file sizes in the supplied directory on the supplied host by ssh (or on localhost if no host is supplied)
    """
    # determine file sizes with du
    return run_process('"(cd ' + dir_name + ' && du -sm *)"', process_du, host)

def get_status(host=None, dir_name=None):
    """Base entry point for getting all status for a host, returned as a dict
    Arguments:
      host - optional remote host (must be accessible via ssh)
      dir_name - directory to check disk space on
    """
    # determine file sizes with du
    status = {}
    if(host != None):
        status['host'] = host
    if(dir_name != None):
        status['dir'] = dir_name
        status.update(run_process('df -P -BG ' + dir_name, process_df, host))
    status.update(run_process('free -m', process_free, host))
    status.update(run_process('uptime', process_uptime, host))
    status.update(run_process('grep -c "^processor" /proc/cpuinfo', process_ncores, host))
    return status

up_pattern = re.compile('.* load average: ([0-9.]+), ([0-9.]+), ([0-9.]+)')


def process_uptime(status, line):
    """Internal method to parse output of uptime and add to status hash"""
    m = up_pattern.match(line)
    if m:
        status['load_1m'] = float(m.group(1))
        status['load_5m'] = float(m.group(2))
        status['load_15m'] = float(m.group(3))


def process_free(status, line):
    """Internal method to parse output of free and add to status hash"""
    if line.startswith("Mem:"):
        elems = line.split()
        # memory is added in Mb
        status['memory_total_m'] = int(elems[1])
        status['memory_used_m'] = int(elems[2])
        status['memory_available_m'] = int(elems[3])
        status['memory_used_pct'] = float(format(100.0 * status['memory_used_m'] / status['memory_total_m'], '.1f'))


def process_df(status, line):
    """Internal method to parse output of df and add to status hash"""
    if not line.startswith("Filesystem"):
        elems = line.split()
        # space is added in Gb
        status['disk_total_g'] = int(elems[1].replace('G', ''))
        status['disk_used_g'] = int(elems[2].replace('G', ''))
        status['disk_available_g'] = int(elems[3].replace('G', ''))
        status['disk_used_pct'] = float(format(100.0 * status['disk_used_g'] / status['disk_total_g'], '.1f'))


def process_ncores(status, line):
    """Internal method to parse output of cpuinfo file and add to status hash"""
    elems = line.split()
    if elems:
        status['n_cpus'] = int(elems[0])


def process_du(status, line):
    """Internal method to parse output of du and add to status hash"""
    elems = line.split()
    if elems:
        status[elems[1]] = int(elems[0])


def run_process(command, function, host=None):
    """Internal common method used to execute supplied command on supplied host, and parse with supplied function"""
    status = {}
    if host != None:
        command = ['ssh -q', host, command]
    proc = Popen(command, shell=True, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    try:
        outs, errs = proc.communicate(timeout=20)
    except TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()
    if proc.returncode != 0:
        status['command_error'] = {'stdout': outs, 'stderr': errs}
    for line in outs.splitlines():
        function(status, line)
    return status




# See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import ssl

import urllib3
from elasticsearch import Elasticsearch, ElasticsearchException
from elasticsearch.connection import create_ssl_context


class ElasticsearchConnectionManager:
    def __init__(self, host: str, port: int, user: str = "", password: str = "", with_ssl: bool = False):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.ssl = with_ssl
        self.client = None

    def __enter__(self):
        urllib3.disable_warnings(category=urllib3.connectionpool.InsecureRequestWarning)
        ssl_context = create_ssl_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
        self.client = Elasticsearch(hosts=[{'host': self.host, 'port': self.port}],
                                    scheme="https" if self.ssl else "http",
                                    ssl_context=ssl_context,
                                    http_auth=(self.user, self.password))
        if not self.client.ping():
            raise RuntimeError(
                f"Cannot connect to Elasticsearch server. User: {self.user}, Host: {self.host}:{self.port}")
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.client.transport.close()



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from urllib.parse import urlparse
from sqlalchemy import create_engine, text
from ensembl.production.core.server_utils import get_file_sizes
from sqlalchemy.engine.url import make_url


def validate_mysql_url(db_uri):
    parsed_uri = urlparse(db_uri)
    if parsed_uri.scheme != 'mysql':
        raise ValueError('list_databases can only work with MySQL databases')
    try:
        if parsed_uri.port is None:
            raise ValueError('Invalid port number in db_uri')
    except ValueError as e:
        raise ValueError('Invalid port: {}'.format(e))
    return db_uri


def list_databases(db_uri, query):
    """
    List databases on a specified MySQL server
    Arguments:
      db_uri : URI of MySQL server e.g. mysql://user@host:3306/
      query : optional regular expression to filter databases e.g. .*_core_.*
    """
    valid_uri = validate_mysql_url(db_uri)
    engine = create_engine(valid_uri)
    if(query == None):
        s = text("select schema_name from information_schema.schemata")
    else:
        s = text("select schema_name from information_schema.schemata where schema_name rlike :q")
    with engine.connect() as con:
        return [str(r[0]) for r in con.execute(s, q=query).fetchall()]


def get_database_sizes(db_uri, query, dir_name):
    """
    List sizes of databases on a specified MySQL server
    Arguments:
      db_uri : URI of MySQL server e.g. mysql://user@host:3306/ (file system must be accessible)
      query : optional regular expression to filter databases e.g. .*_core_.*
      dir_name : location of MySQL data files on server
    """
    db_list = list_databases(db_uri, query)
    url = make_url(db_uri)
    dir_path = os.path.join(dir_name, str(url.port), 'data')
    sizes = get_file_sizes(url.host, dir_path)
    return {db: sizes[db] for db in db_list if db in sizes.keys()}


def get_databases_list(server_uri, query):
    """
    List databases on a specified MySQL server
    Arguments:
      server_uri : URI of MySQL server e.g. mysql://user@host:3306/
      query : optional regular expression to filter databases e.g. .*_core_.*
    """
    if server_uri.startswith('mysql') is False:
        raise ValueError('list_databases can only work with MySQL databases')

    engine = create_engine(server_uri)
    if query is None:
        s = text("select schema_name from information_schema.schemata")
    else:
        s = text("select schema_name from information_schema.schemata where schema_name rlike :q")
    with engine.connect() as con:
        return [str(r[0]) for r in con.execute(s, {"q": query}).fetchall()]


def get_db_type(db_uri):
    """
    Retrieve the value for the 'schema_type' meta_key from a database
    Arguments:
      db_uri : URI of MySQL database e.g. mysql://user@host:3306/db_name_core_100_1
    """
    s = text("select meta_value from meta where meta_key = 'schema_type'")
    engine = create_engine(db_uri)
    with engine.connect() as con:
        row = con.execute(s).fetchone()
        return str(row[0])



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import unittest

from ensembl.production.core.perl_utils import dict_to_perl_string, perl_string_to_python

logging.basicConfig()


class UtilsTest(unittest.TestCase):

    def test_parse_simple_str(self):
        o = {'a': 'b'}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => "b"}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(o2['a'], 'b')

    def test_parse_simple_str_pair(self):
        o = {'a': 'b', 'c': 'd'}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => "b", "c" => "d"}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(o2['a'], 'b')
        self.assertEqual(o2['c'], 'd')

    def test_parse_simple_int(self):
        o = {'a': 99}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => 99}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(o2['a'], 99)

    def test_parse_simple_list(self):
        o = {'a': ["a", "b", "c"]}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => ["a", "b", "c"]}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(type(o2['a']).__name__, "list")

    def test_parse_complex(self):
        o = {'a': ["a", "b", "c"], 'b': 'd', 'c': 99}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => ["a", "b", "c"], "b" => "d", "c" => 99}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(type(o2['a']).__name__, "list")
        self.assertEqual(o2['b'], 'd')
        self.assertEqual(o2['c'], 99)

    def test_parse_string_quotes(self):
        o = {'a': "\"b\""}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => "\\"b\\""}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(o2['a'], '"b"')

    def test_parse_string_dollar(self):
        o = {'a': "$b"}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => "\\$b"}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(o2['a'], '$b')

    def test_parse_string_at(self):
        o = {'a': "@b"}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{"a" => "\\@b"}""")
        o2 = perl_string_to_python(s)
        t = type(o2).__name__
        self.assertEqual(t, "dict")
        self.assertEqual(o2['a'], '@b')

    def test_parse_none(self):
        o = {'a': None}
        s = dict_to_perl_string(o)
        self.assertEqual(s, """{}""")



#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ensembl.production.core.rest import RestClient
from ensembl.production.core.server_utils import assert_mysql_uri
import re
import json
import logging


class DatacheckClient(RestClient):
    """Client for checking databases using the datacheck service"""

    def submit_job(self, server_url, dbname, species, division, db_type,
                   datacheck_names, datacheck_groups, datacheck_types,
                   email, tag, target_url):
        #FIXME Make signature identical to parent
        """
    Run datachecks on a given server, for one or more species.
    Parameter requirements are complicated, because only the server_url is absolutely required,
    for lots of other parameters you need one from a set, but it doesn't matter which one...
    Arguments:
      server_url - location of server, in URI format
      dbname - name of a database to check
      species - name of a species to check
      division - name of a division to check
      db_type - type of database to check, defaults to 'core'

      datacheck_names - names of datacheck(s) to run, multiple values must be comma-separated
      datacheck_groups - datacheck group(s) to run, multiple values must be comma-separated
      datacheck_types - optional filter on type, 'critical' or 'advisory'

      email - optional address for an email on job completion
      tag - optional text for grouping datacheck submissions
      target_url - optional location of 'ancillary' server, for related databases
    """
        assert_mysql_uri(server_url)

        payload = {
            'server_url': server_url,
            'dbname': dbname,
            'species': species,
            'division': division,
            'db_type': db_type,
            'datacheck_names': [],
            'datacheck_groups': [],
            'datacheck_types': [],
            'email': email,
            'tag': tag,
        }

        if target_url is not None:
            payload['target_url'] = target_url
        if datacheck_names is not None:
            payload['datacheck_names'] = datacheck_names.split(',')
        if datacheck_groups is not None:
            payload['datacheck_groups'] = datacheck_groups.split(',')
        if datacheck_types is not None:
            payload['datacheck_types'] = datacheck_types.split(',')

        return RestClient.submit_job(self, payload)

    def list_jobs(self, output_file, pattern, failure_only=False):
        """
    Find jobs and print results
    Arguments:
      output_file - optional file to write report
      pattern - optional pattern to filter jobs by
      failure_only - only report failed jobs
    """
        jobs = super(DatacheckClient, self).list_jobs()
        if pattern is None:
            pattern = '.*'
        tag_pattern = re.compile(pattern)
        output = []
        for job in jobs:
            if 'tag' in job['input']:
                tag = job['input']['tag']
            else:
                tag = ''
            if tag_pattern.search(tag):
                if 'output' in job:
                    if failure_only is True:
                        if job['output']['failed_total'] > 0:
                            output.append(job)
                    else:
                        output.append(job)
                else:
                    output.append(job)

        if output_file is None:
            print(json.dumps(output, indent=2))
        else:
            output_file.write(json.dumps(output))

    def print_job(self, job, print_results=False, print_input=False):
        """
    Render a job to logging
    Arguments:
      job :  job to print
      print_results : set to True to print detailed results
      print_input : set to True to print input for job
    """
        logging.info("Job %s - %s" % (job['id'], job['status']))
        if print_input is True:
            self.print_inputs(job['input'])
        if job['status'] == 'complete':
            if print_results is True:
                logging.info("Submission status: " + str(job['status']))
                logging.info("Database passed: " + str(job['output']['passed_total']))
                logging.info("Database failed: " + str(job['output']['failed_total']))
                logging.info("Output directory: " + str(job['output']['output_dir']))
                logging.info("Per database results: ")
                logging.info(json.dumps(job['output']['databases'], indent=2))
        elif job['status'] == 'incomplete':
            if print_results is True:
                logging.info("Submission status: " + str(job['status']))
        elif job['status'] == 'failed':
            logging.info("Submission status: " + str(job['status']))
            # failures = self.retrieve_job_failure(job['id'])
            # logging.info("Error: " + str(failures))
        else:
            raise ValueError("Unknown status {}".format(job['status']))

    def print_inputs(self, i):
        """Utility to render a job input dict to logging"""
        logging.info("Registry file: " + i['registry_file'])
        if 'dbname' in i:
            for dbname in i['dbname']:
                logging.info("Database name: " + dbname)
        if 'species' in i:
            for species in i['species']:
                logging.info("Species name: " + species)
        if 'division' in i:
            for division in i['division']:
                logging.info("Division name: " + division)
        if 'db_type' in i:
            logging.info("Database type: " + i['db_type'])
        if 'datacheck_names' in i:
            for name in i['datacheck_names']:
                logging.info("Datacheck: " + name)
        if 'datacheck_groups' in i:
            for group in i['datacheck_groups']:
                logging.info("Datacheck group: " + group)
        if 'datacheck_types' in i:
            for datacheck_type in i['datacheck_types']:
                logging.info("Datacheck type: " + datacheck_type)
        if 'email' in i:
            logging.info("Email: " + i['email'])
        if 'tag' in i:
            logging.info("Tag: " + i['tag'])



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import re

import time
from sqlalchemy import create_engine, Column, Integer, String, ForeignKey, func, and_
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, joinedload

from ensembl.production.core.perl_utils import dict_to_perl_string, perl_string_to_python

__all__ = ['Result', 'LogMessage', 'Job', 'HiveInstance', 'Analysis']

Base = declarative_base()

logger = logging.getLogger(__name__)


class Analysis(Base):
    __tablename__ = 'analysis_base'

    analysis_id = Column(Integer, primary_key=True)
    logic_name = Column(String)

    def __repr__(self):
        return "<Analysis(analysis_id='%s', logic_name='%s')>" % (
            self.analysis_id, self.logic_name)


class AnalysisData(Base):
    __tablename__ = 'analysis_data'

    analysis_data_id = Column(Integer, primary_key=True)
    data = Column(String)

    def __repr__(self):
        return "<AnalysisData(analysis_data_id='%s', data='%s')>" % (
            self.analysis_data_id, self.data)


class Beekeeper(Base):
    __tablename__ = 'beekeeper'

    beekeeper_id = Column(Integer, primary_key=True)
    cause_of_death = Column(String)
    loop_until = Column(String)

    def __repr__(self):
        return "<Beekeeper(beekeeper_id='%s', cause_of_death='%s')>" % (
            self.beekeeper_id, self.cause_of_death)


class Result(Base):
    __tablename__ = 'result'

    job_id = Column(Integer, primary_key=True)
    output = Column(String)

    def output_dict(self):
        return json.loads(self.output)

    def __repr__(self):
        return "<Result(job_id='%s', output='%s')>" % (
            self.job_id, self.output)


class JobProgress(Base):
    __tablename__ = 'job_progress'

    job_progress_id = Column(Integer, primary_key=True)
    job_id = Column(Integer)
    message = Column(String)

    def __repr__(self):
        return "<JobProgress(job_progress_id='%s', job_id='%s', message='%s')>" % (
            self.job_progress_id, self.job_id, self.message)


class LogMessage(Base):
    __tablename__ = 'log_message'

    log_message_id = Column(Integer, primary_key=True)
    job_id = Column(Integer)
    msg = Column(String)
    status = Column(String)
    message_class = Column(String)
    when_logged = Column(String)

    def __repr__(self):
        return "<LogMessage(log_message_id='%s', msg='%s')>" % (
            self.log_message_id, self.msg)


class Role(Base):
    __tablename__ = 'role'

    role_id = Column(Integer, primary_key=True)
    worker_id = Column(Integer, ForeignKey("worker.worker_id"))

    def __repr__(self):
        return "<Role(role_id='%s', worker_id='%s')>" % (
            self.role_id, self.worker_id)


class Worker(Base):
    __tablename__ = 'worker'

    worker_id = Column(Integer, primary_key=True)
    process_id = Column(String)

    def __repr__(self):
        return "<Worker(worker_id='%s', process_id='%s')>" % (
            self.worker_id, self.process_id)


class Semaphore(Base):
    __tablename__ = 'semaphore'

    semaphore_id = Column(Integer, primary_key=True)
    dependent_job_id = Column(Integer)
    local_jobs_counter = Column(Integer, default=0)

    def __repr__(self):
        return "<Semaphore(semaphore_id = '%s', dependent_job_id='%s', local_jobs_counter='%s')>" % (
            self.semaphore_id, self.dependent_job_id, self.local_jobs_counter)


class Job(Base):
    __tablename__ = 'job'

    job_id = Column(Integer(), ForeignKey("result.job_id"), ForeignKey("log_message.job_id"), primary_key=True,
                    autoincrement=True)
    input_id = Column(String)
    status = Column(String)
    param_id_stack = Column(String, default='')
    prev_job_id = Column(Integer)
    controlled_semaphore_id = Column(Integer, ForeignKey("semaphore.semaphore_id"))
    role_id = Column(Integer, ForeignKey("role.role_id"))

    analysis_id = Column(Integer, ForeignKey("analysis_base.analysis_id"))
    analysis = relationship("Analysis", uselist=False, lazy="joined")

    result = relationship("Result", uselist=False, lazy="joined")

    log_messages = relationship("LogMessage", viewonly=True)
    when_completed = Column(String)

    def __repr__(self):
        return "<Job(job_id='%s', analysis='%s', input_id='%s', status='%s', result='%s', role=%s, when_completed=%s)>" % (
            self.job_id, self.analysis.logic_name, self.input_id, self.status,
            (self.result.output if self.result is not None else None),
            self.role_id if self.result is not None else None,
            self.when_completed)


Session = sessionmaker()


class HiveInstance:
    analysis_dict = {}

    def __init__(self, url, timeout=3600):
        self.engine = create_engine(url, pool_recycle=timeout)
        Session.configure(bind=self.engine)

    def get_job_by_id(self, id):
        """ Retrieve a job given the unique surrogate ID """
        with Session() as session:
            return self._get_job_by_id(id, session)

    def _get_job_by_id(self, id, session):
        job = session.query(Job).options(joinedload(Job.result)).filter(Job.job_id == id).first()
        if job is None:
            raise ValueError("Job %s not found" % id)
        return job

    def get_worker_id(self, id):
        """ Retrieve a worker_id for a given role_id """
        with Session() as session:
            return self._get_worker_id(id, session)

    def _get_worker_id(self, id, session):
        return session.query(Role).filter(Role.role_id == id).first()

    def get_jobs_failure_msg(self, id):
        """Get failures for all the parent and child jobs"""
        with Session() as session:
            return self._get_jobs_failure_msg(id, session)

    def _get_jobs_failure_msg(self, id, session):
        failures = {}
        parent_job = self.get_job_by_id(id)
        if parent_job.status == 'FAILED':
            failures[id] = self._get_job_failure_msg_by_id(id, session).msg
        for child_job in session.query(Job).filter(Job.prev_job_id == id).all():
            if child_job.status == 'FAILED':
                failures[child_job.job_id] = self._get_job_failure_msg_by_id(child_job.job_id, session).msg
        return failures

    def get_job_failure_msg_by_id(self, id, child=False):
        """ Retrieve a job failure message or job child if exist and if child flag turned on"""
        with Session() as session:
            return self._get_job_failure_msg_by_id(id, session, child)

    def _get_job_failure_msg_by_id(self, id, session, child=False):
        job = self._get_job_by_id(id, session)
        if job is None:
            raise ValueError("Job %s not found" % id)
        if child:
            child_job = self._get_job_child(job, session)
            if child_job is not None:
                return session.query(LogMessage).filter(LogMessage.job_id == child_job.job_id).order_by(
                    LogMessage.log_message_id.desc()).first()
            return session.query(LogMessage).filter(LogMessage.job_id == id).order_by(
                LogMessage.log_message_id.desc()).first()
        return session.query(LogMessage).filter(LogMessage.job_id == id).order_by(
            LogMessage.log_message_id.desc()).first()

    def get_worker_process_id(self, id):
        """ Find a workers process_id """
        with Session() as session:
            return self._get_worker_process_id(id, session)

    def _get_worker_process_id(self, id, session):
        return session.query(Worker).filter(Worker.worker_id == id).first()

    def get_analysis_by_name(self, name):
        """ Find an analysis """
        with Session() as session:
            return self._get_analysis_by_name(name, session)

    def _get_analysis_by_name(self, name, session):
        return session.query(Analysis).filter(Analysis.logic_name == name).first()

    def create_job(self, analysis_name, input_data):
        """
        Create a job for the supplied analysis and input hash
        The input_data dict is converted to a Perl string before storing
        """
        with Session() as session:
            return self._create_job(analysis_name, input_data, session)

    def _create_job(self, analysis_name, input_data, session):
        input_data['timestamp'] = time.ctime()
        analysis = self._get_analysis_by_name(analysis_name, session)
        if analysis is None:
            raise ValueError("Analysis %s not found" % analysis_name)
        job = Job(input_id=dict_to_perl_string(input_data), status='READY', analysis_id=analysis.analysis_id)
        session.add(job)
        session.commit()
        # force load of object
        # FIXME any param in connexion should force the loading.
        job.analysis
        job.result
        return job

    def get_analysis_data_input(self, analysis_data_id):
        """ Get the job input stored in the analysis_data table. Get input from child job if exist"""
        with Session() as session:
            return self._get_analysis_data_input(analysis_data_id, session)

    def _get_analysis_data_input(self, analysis_data_id, session):
        return session.query(AnalysisData).filter(AnalysisData.analysis_data_id == analysis_data_id).first()

    def get_semaphore_data(self, semaphore_job_id):
        """ Get the job semaphore count if exist"""
        with Session() as session:
            return self._get_semaphore_data(semaphore_job_id, session)

    def _get_semaphore_data(self, semaphore_job_id, session):
        return session.query(Semaphore).filter(Semaphore.dependent_job_id == semaphore_job_id).first()

    def get_result_for_job_id(self, id, child=False, progress=True, analysis_id=None):
        """ Get result for a given job id. If child flag is turned on and job child exist, get result for child job"""
        with Session() as session:
            return self._get_result_for_job_id(id, session, child, progress, analysis_id)

    def _get_result_for_job_id(self, id, session, child=False, progress=True, analysis_id=None):
        job = self._get_job_by_id(id, session)
        if job is None:
            raise ValueError(f"Job {id} not found")
        if child:
            child_job = self._get_job_child(job, session)
            if child_job is not None:
                return self._get_result_for_job(child_job, session, progress)
            return self._get_result_for_job(job, session, progress)
        return self._get_result_for_job(job, session, progress, analysis_id)

    def get_result_for_job(self, job, progress=False, analysis_id=None):
        """
        Determine if the job has completed. If the job has semaphored children, they are also checked
        Also return progress of jobs, completed and total if flag is on
        """
        with Session() as session:
            return self._get_result_for_job(job, session, progress, analysis_id)

    def _get_result_for_job(self, job, session, progress=False, analysis_id=None):
        result = {"id": job.job_id}
        try:
            if re.search(r"^(_extended_data_id){1}(\s){1}(\d+){1}", job.input_id):
                extended_data = job.input_id.split(" ")
                job_input = self._get_analysis_data_input(extended_data[1], session)
                result['input'] = perl_string_to_python(job_input.data)

            else:
                result['input'] = perl_string_to_python(job.input_id)
            if job.status == 'DONE' and job.result is not None:
                result['status'] = 'complete'
                result['when_completed'] = job.when_completed
                result['output'] = job.result.output_dict()
            else:
                result['status'] = self._get_job_tree_status(job, session)
            if progress:
                result['progress'] = self._get_all_jobs_progress(job.job_id, session, analysis_id=analysis_id)
        except ValueError as e:
            raise ValueError(f'Cannot retrieve results for job: {job.job_id}') from e
        except SQLAlchemyError as e:
            raise ValueError(f'DB error for job: {job.job_id}') from e
        return result

    def get_all_jobs_progress(self, job_id, analysis_id=None):
        """
        Get all jobs from Job table based on given job id and analysis id
        """
        with Session() as session:
            return self._get_all_jobs_progress(job_id, session, analysis_id)

    def _get_all_jobs_progress(self, job_id, session, analysis_id=None):
        results = {'total': 0, 'inprogress': 0, 'completed': 0, 'failed': 0 }
        job_pattern = f"{job_id},%"

        if analysis_id:
            jobs = session.query(Job).filter(and_(Job.param_id_stack.ilike(job_pattern), Job.analysis_id == analysis_id ) ).all()
        else:
            jobs = session.query(Job).filter(Job.param_id_stack.ilike(job_pattern) ).all()

        for job in jobs:
            results['total'] += 1
            if job.status == 'DONE':
                results['completed'] += 1
            elif job.status == 'FAILED':
                results['failed'] += 1
            else:
                results['inprogress'] += 1
        return results

    def get_last_job_progress(self, job):
        """ Return last job progress line if exists, else None """
        with Session() as session:
            return self._get_last_job_progress(job, session)

    def _get_last_job_progress(self, job, session):
        last_job_progress_msg = session.query(JobProgress).filter(JobProgress.job_id == job.job_id).order_by(
            JobProgress.job_progress_id.desc()).first()
        return last_job_progress_msg

    def get_jobs_progress(self, job):
        """
        Check data in the job_progress table
        alternatively, get jobs progress for parent and children jobs
        Return number of completed jobs and total of jobs
        If there is data in the job_progress table, return progress message
        """
        with Session() as session:
            return self._get_jobs_progress(job, session)

    def _get_jobs_progress(self, job, session):
        last_job_progress_msg = session.query(JobProgress).filter(JobProgress.job_id == job.job_id).order_by(
            JobProgress.job_progress_id.desc()).first()
        if last_job_progress_msg is not None:
            total = 10
            complete = session.query(JobProgress).filter(JobProgress.job_id == job.job_id).count()
            return {"complete": complete, "total": total, "message": last_job_progress_msg.message}
        total = 1
        complete = 0
        parent_job = self._get_job_by_id(job.job_id, session)
        if parent_job.status == 'DONE':
            complete += 1
        for child_job in session.query(Job).filter(Job.prev_job_id == job.job_id).all():
            total += 1
            if child_job.status == 'DONE':
                complete += 1
        return {"complete": complete, "total": total}

    def get_job_tree_status(self, job):
        """ Recursively check all children of a job """
        with Session() as session:
            return self._get_job_tree_status(job, session)

    def _get_job_tree_status(self, job, session):
        # check for semaphores
        semaphore_data = None
        logger.debug("get_job_tree_status :: job: %s", job)
        semaphored_job = session.query(Job).filter(Job.prev_job_id == job.job_id and job.status == 'SEMAPHORED').first()
        logger.debug("get_job_tree_status :: semaphored_job: %s", semaphored_job)
        if semaphored_job is not None:
            semaphore_data = self._get_semaphore_data(semaphored_job.job_id, session)
        logger.debug("get_job_tree_status :: semaphore_data: %s", semaphore_data)
        if semaphore_data is not None and semaphore_data.local_jobs_counter > 0:
            return self._check_semaphores_for_job(semaphore_data, session)
        if job.status == 'FAILED':
            return 'failed'
        if job.status == 'READY':
            return 'submitted'
        if job.status == 'RUN':
            return 'running'
        if job.status == 'DONE':
            for child_job in session.query(Job).filter(Job.prev_job_id == job.job_id).all():
                child_status = self._get_job_tree_status(child_job, session)
                if child_status != 'complete':
                    return child_status
            return 'complete'
        return 'incomplete'

    def get_job_child(self, job):
        """ Get child job for a given parent job """
        with Session() as session:
            return self._get_job_child(job, session)

    def _get_job_child(self, job, session):
        return session.query(Job).filter(Job.prev_job_id == job.job_id).first()

    def get_job_parent(self, job):
        """ Get parent job for a given children job """
        with Session() as session:
            return self._get_job_parent(job, session)

    def _get_job_parent(self, job, session):
        return session.query(Job).filter(Job.job_id == job.prev_job_id).first()

    def get_semaphored_jobs(self, job, status=None):
        """
        Find all jobs that are semaphored children of the nominated job, optional filtering by status
        'complete' indicates that all children completed successfully
        'failed' indicates that at least one child has failed
        'incomplete' indicates that at least one child is running or ready
        """
        with Session() as session:
            return self._get_semaphored_jobs(job, session, status)

    def _get_semaphored_jobs(self, job, session, status=None):
        semaphored_job = session.query(Job).filter(Job.prev_job_id == job.job_id, job.status == 'SEMAPHORED').first()
        semaphore_data = self._get_semaphore_data(semaphored_job.job_id, session)
        if status is None:
            return session.query(Job).filter(semaphore_data.semaphore_id == Job.controlled_semaphore_id).all()
        return session.query(Job).filter(semaphore_data.semaphore_id == Job.controlled_semaphore_id, Job.status == status).all()

    def check_semaphores_for_job(self, semaphore_data):
        """ Find all jobs that are semaphored children of the nominated job, and check whether they have completed """
        with Session() as session:
            return self._check_semaphores_for_job(semaphore_data, session)

    def _check_semaphores_for_job(self, semaphore_data, session):
        status = 'complete'
        jobs = dict(session.query(Job.status, func.count(Job.status)).filter(
            semaphore_data.semaphore_id == Job.controlled_semaphore_id).group_by(Job.status).all())
        logger.debug("check_semaphores_for_job :: jobs: %s", jobs)
        if jobs.get('FAILED', 0) > 0:
            status = 'failed'
        elif jobs.get('READY', 0) > 0 or jobs.get('RUN', 0) > 0 or jobs.get('SEMAPHORED', 0) > 0:
            status = 'incomplete'
        return status

    def get_all_results(self, analysis_name, child=False):
        """Find all jobs from the specified analysis"""
        with Session() as session:
            return self._get_all_results(analysis_name, session, child)

    def _get_all_results(self, analysis_name, session, child=False):
        jobs = session.query(Job).join(Analysis).filter(Analysis.logic_name == analysis_name).all()
        if child:
            return list(
                map(lambda job: self._get_result_for_job(self._get_job_child(job, session), session) if (
                    self._get_job_child(job, session) is not None) else self._get_result_for_job(job, session),
                    jobs)
            )
        return list(map(lambda job: self._get_result_for_job(job, session), jobs))

    def delete_job_by_id(self, job_id, child=False):
        """Delete a job from the hive database given its id"""
        with Session() as session:
            return self._delete_job_by_id(job_id, session, child)

    def _delete_job_by_id(self, job_id, session, child=False):
        job = self._get_job_by_id(job_id, session)
        return self._delete_job(job, session, child)

    def delete_job(self, job, child=False):
        """Delete a job from the hive database
           If child flag turn on, try to delete child job if exist
           Also get parent job if exist and delete it """
        with Session() as session:
            return self._delete_job(job, session, child)

    def _delete_job(self, job, session, child=False):
        parent_job = self._get_job_parent(job, session)
        if child:
            child_job = self._get_job_child(job, session)
            if child_job is not None:
                logger.debug("Deleting children job %s", child_job.job_id)
                if child_job.result is not None:
                    session.delete(child_job.result)
                session.delete(child_job)
                session.commit()
        if parent_job is not None:
            logger.debug("Deleting parent job %s", parent_job.job_id)
            if parent_job.result is not None:
                session.delete(parent_job.result)
            session.delete(parent_job)
            session.commit()
        logger.debug("Deleting job %s", job.job_id)
        if job.result is not None:
            session.delete(job.result)
        session.delete(job)
        session.commit()



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

class APIException(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload
        Exception.__init__(self, self.message, self.status_code)

    def __str__(self):
        return self.message

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


class HTTPRequestError(APIException):
    pass




# See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import unittest

from ensembl.production.core.db_utils import validate_mysql_url


class TestPerlParsing(unittest.TestCase):

    def test_extra_hidden_character(self):
        """
        Use case tested for parsing output from metadata updated, ensure the hidden chars \t\r\n are actually
        removed from initial string.
        :return: None
        """
        from ensembl.production.core.perl_utils import perl_string_to_python
        test_string = '{"comment" => "Handover 105 core of new Metazoan species database agrilus	planipennis", ' \
                      '"database_uri" => "mysql://ensuser:enspass\@localhost:3306/agrilus_planipennis_' \
                      'gca000699045v2_core_105_1", "email" => "test\@ebi.ac.uk", "metadata_uri" => ' \
                      '"mysql://ensuser:enspass\@localhost:3306/ensembl_metadata_qrp", ' \
                      '"source" => "Handover", "timestamp" => "Thu Jan 13 17:45:37 2022"}'

        data_python = perl_string_to_python(test_string)
        self.assertEqual(len(data_python.keys()), 6)
        self.assertIn('comment', data_python.keys())
        self.assertNotIn("\t", data_python['comment'])
        db_uri = validate_mysql_url(data_python['database_uri'])
        self.assertEqual(db_uri, 'mysql://ensuser:enspass@localhost:3306/agrilus_planipennis_gca000699045v2_core_105_1')

    def test_compat_import(self):
        from ensembl.production.core.utils import perl_string_to_python
        perl_string = '{"testKey" => "testValue\@testhost"}'
        with self.assertWarns(DeprecationWarning):
            data_python = perl_string_to_python(perl_string)
            self.assertEqual(len(data_python.keys()), 1)
            self.assertEqual(data_python['testKey'], 'testValue@testhost')



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import unittest

from ensembl.production.core.es import ElasticsearchConnectionManager


# TODO ADD more test with proper config in gitlab
class EsTest(unittest.TestCase):
    def test_es_exception(self):
        with self.assertRaises(RuntimeError) as expected:
            with ElasticsearchConnectionManager(host="http://unknown",
                                                port=9200) as elastic:
                elastic.client.ping()
            self.assertTrue('Cannot connect to Elasticsearch server' in expected)




#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import re
import os
import logging
import yaml

logger = logging.getLogger(__name__)


def load_config_yaml(file_path, strict=False, tag='!ENV'):
    """
    Load a yaml configuration file and resolve any environment variables
    The environment variables must have !ENV before them and be in this format
    to be parsed: ${VAR_NAME}.
    E.g.:
    database:
        host: !ENV ${HOST}
        port: !ENV ${PORT}
    app:
        log_path: !ENV '/var/${LOG_PATH}'
        something_else: !ENV '${AWESOME_ENV_VAR}/var/${A_SECOND_AWESOME_VAR}'
    :param strict: String yaml loading
    :param str file_path: the path to the yaml file
    :param str tag: the tag to look for
    :return: the dict configuration
    :rtype: dict[str, T]
    See https://gist.github.com/mkaranasou/ba83e25c835a8f7629e34dd7ede01931#file-python_yaml_environment_variables-py
    """
    pattern = re.compile(r'.*?\${(\w+)}.*?')
    loader = yaml.SafeLoader

    # the tag will be used to mark where to start searching for the pattern
    # e.g. somekey: !ENV somestring${MYENVVAR}blah blah blah
    loader.add_implicit_resolver(tag, pattern, None)

    def constructor_env_variables(loader, node):
        """
        Extracts the environment variable from the node's value
        :param yaml.Loader loader: the yaml loader
        :param node: the current node in the yaml
        :return: the parsed string that contains the value of the environment
        variable
        """
        value = loader.construct_scalar(node)
        match = pattern.findall(value)  # to find all env variables in line
        if match:
            full_value = value
            for g in match:
                full_value = full_value.replace(
                    f'${{{g}}}', os.environ.get(g, g)
                )
            return full_value
        return value

    loader.add_constructor(tag, constructor_env_variables)

    if file_path:
        with open(file_path, 'r') as f:
            config = yaml.load(f, Loader=loader)
        return config if config else {}
    else:
        if strict:
            raise ValueError('Invalid config file path: %s' % file_path)
        else:
            logger.warning('Using default configuration. Config file path was: %s', file_path)
            return {}


def parse_debug_var(var):
    return not ((str(var).lower() in ('f', 'false', 'no', 'none')) or (not var))


def load_config_json(file_path):
    with open(file_path, 'r') as f:
        config = json.load(f)
    return config if config else {}



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import datetime
import logging

from sqlalchemy import create_engine, Column, Integer, String, DateTime, ForeignKey, Enum
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship

logging.basicConfig()


def lazy_load(obj):
    """
    Helper method to call all attribs on an obj to load them before detaching from the session
    """
    [getattr(obj, method) for method in dir(obj) if callable(getattr(obj, method))]


Base = declarative_base()


class LockEnum(Enum):
    read = 1
    write = 2


class ResourceLock(Base):
    """Class respresenting a lock obtained by a client on a particular resource. Locks can include read or write.

    Attributes:
    resource_lock_id -- ID for the lock (given to the calling code for release)
    lock_type        -- read or write
    created          -- time at which lock was obtained
    client           -- Client holding the lock
    resource         -- Resource being locked
    """
    __tablename__ = 'resource_lock'

    resource_lock_id = Column(Integer, primary_key=True)
    lock_type = Column(String(5), nullable=False)
    created = Column(DateTime, nullable=False, default=datetime.datetime.utcnow)
    client_id = Column(Integer, ForeignKey("client.client_id"))
    client = relationship("Client")
    resource_id = Column(Integer, ForeignKey("resource.resource_id"))
    resource = relationship("Resource")

    def __repr__(self):
        return "<ResourceLock(resource_lock_id={}, lock_type='{}', client='{}', resource='{}')>".format(
            self.resource_lock_id, self.lock_type, self.client.name, self.resource.uri)

    def to_dict(self):
        return {"resource_lock_id": self.resource_lock_id, "client": self.client.to_dict(),
                "resource": self.resource.to_dict(), "lock_type": self.lock_type, "created": self.created}


class Resource(Base):
    """Class respresenting an abstract resource like a database or a file

    Attributes:
    resource_id -- internal ID for the resource
    uri         -- unique string representation of the resource e.g. database URI, path to file
    """
    __tablename__ = 'resource'

    resource_id = Column(Integer, primary_key=True)
    uri = Column(String(512), nullable=False, unique=True)

    def __repr__(self):
        return "<Resource(resource_id={}, uri='{}')>".format(
            self.resource_id, self.uri)

    def to_dict(self):
        return {"resource_id": self.resource_id, "uri": self.uri}


class Client(Base):
    """Class respresenting an abstract client who needs access to a resource.
    A client might be a process, application or Real Person[tm]

    Attributes:
    client_id -- internal ID for the client
    name      -- unique string name for client. Could be the name of an application, or an email address for a person
    """

    __tablename__ = 'client'

    client_id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False, unique=True)

    def __repr__(self):
        return "<Client(client_id={}, name='{}')>".format(
            self.client_id, self.name)

    def to_dict(self):
        return {"client_id": self.client_id, "name": self.name}


class LockException(Exception):
    pass


Session = sessionmaker()


class ResourceLocker:
    """Utility class for the locking and unlocking of resources
    """

    def __init__(self, url, timeout=3600):
        """Create a new ResourceLocker instance
        Attributes:
          url - URL of backing database
          timeout - (optional) time in seconds to keep connections to database open. Defaults to 3600s
        """
        self.url = url
        engine = create_engine(url, pool_recycle=timeout, echo=False)
        Base.metadata.create_all(engine)
        Session.configure(bind=engine)

    def get_client(self, name, session=None):
        """Get or create a client with the given name"""
        return self._get_object(Client, name=name, session=session)

    def get_resource(self, uri, session=None):
        """Get or create a resource with the given URI"""
        return self._get_object(Resource, uri=uri, session=session)

    def get_client_by_id(self, client_id):
        """Get client with the specified ID"""
        session = Session()
        try:
            return session.query(Client).filter_by(client_id=client_id).first()
        finally:
            session.close()

    def get_resource_by_id(self, resource_id):
        """Get resource with the specified ID"""
        session = Session()
        try:
            return session.query(Resource).filter_by(resource_id=resource_id).first()
        finally:
            session.close()

    def _get_object(self, obj_type, **kwargs):
        """Fetch or create a basic object from the database.
        If the object is not present, create it.
        If it has already been created, return it.
        If a duplicate exists, retry the method.
        """
        has_session = 'session' in kwargs
        session = kwargs.pop('session', Session())
        if session == None:
            has_session = False
            session = Session()
        try:
            obj = session.query(obj_type).filter_by(**kwargs).first()
            if obj:
                return obj
            else:
                obj = obj_type(**kwargs)
                self._lock_db(session)
                session.add(obj)
                session.commit()
                self._unlock_db(session)
                # lazily load attrs so they can be accessed in a detached object
                lazy_load(obj)
                return obj
        except IntegrityError:
            # duplicate entry, so try again to fetch with a fresh session
            kwargs['session'] = session
            return self._get_object(obj_type, **kwargs)
        finally:
            if has_session == False:
                logging.debug("Closing session")
                session.close()

    def get_locks(self, **kwargs):
        """Fetch current locks from the database
        Optional named arguments for filtering:
          lock_type - read or write
          resource - URI of resource
          client - name of client
        Returns:
          List of ResourceLock objects
        """
        session = Session()
        try:
            q = session.query(ResourceLock)
            resource = kwargs.get('resource_uri')
            client = kwargs.get('client_name')
            lock_type = kwargs.get('lock_type')
            if lock_type != None:
                q = q.filter(ResourceLock.lock_type == lock_type)
            if resource != None:
                q = q.join(Resource).filter(Resource.uri == resource)
            if client != None:
                q = q.join(Client).filter(Client.name == client)
            locks = q.all()
            for l in locks:
                lazy_load(l)
            return locks
        finally:
            session.close()

    def get_lock(self, lock_id):
        """Fetch lock with ID from database
        Argument:
          lock_id - ID of lock
        Returns:
          ResourceLock
        """
        session = Session()
        try:
            lock = session.query(ResourceLock).filter_by(resource_lock_id=lock_id).first()
            lazy_load(lock)
            return lock
        finally:
            session.close()

    def lock(self, client_name, resource_uri, lock_type):
        """Lock the specified resource.
        Arguments:
          client - name of client
          resource - URI of resource
          lock_type - read or write
        Returns:
          ResourceLock
        Raises:
          LockException if resource cannot be locked
          ValueException if lock type not read or write
        """
        logging.info("Locking {} for {} for {}".format(client_name, resource_uri, lock_type))
        session = Session()
        client = self.get_client(client_name, session)
        resource = self.get_resource(resource_uri, session)
        try:
            self._lock_db(session)
            if (lock_type == 'read'):
                # can only create if no write locks found on resource
                n_locks = session.query(ResourceLock).filter_by(resource=resource, lock_type='write').count()
                if (n_locks > 0):
                    raise LockException(
                        "Write lock found on {} - cannot lock for reading {}".format(str(n_locks), resource_uri))
                else:
                    lock = ResourceLock(resource=resource, client=client, lock_type=lock_type)
                    session.add(lock)
                    session.commit()
                    lazy_load(lock)
                    return lock
            elif (lock_type == 'write'):
                # can only create if no other locks found on resource
                n_locks = session.query(ResourceLock).filter_by(resource=resource).count()
                if (n_locks > 0):
                    raise LockException("{} lock(s) found on {}".format(str(n_locks), resource_uri))
                else:
                    lock = ResourceLock(resource=resource, client=client, lock_type=lock_type)
                    session.add(lock)
                    session.commit()
                    lazy_load(lock)
                    return lock
            else:
                raise ValueError("Unsupported lock_type: {}".format(str(lock_type)))
            self._unlock_db(session)
        finally:
            session.close()
        return

    def unlock(self, lock):
        """Release the specified lock
        Arguments:
          lock - either ResourceLock or ID of lock
        Returns:
           None
        Raises:
          ValueError if lock not found
        """
        session = Session()
        if (type(lock) is int):
            lock = session.query(ResourceLock).filter_by(resource_lock_id=lock).first()
            if lock == None:
                raise ValueError("No lock found for ID " + str(lock))
        try:
            logging.info("Deleting lock " + str(lock))
            self._lock_db(session)
            session.delete(lock)
            session.commit()
            self._unlock_db(session)
        finally:
            session.close()
        return

    def get_clients(self):
        """Return all current clients
        Returns:
           List of Client objects
        """
        session = Session()
        try:
            clients = session.query(Client).all()
            for client in clients:
                lazy_load(client)
            return clients
        finally:
            session.close()
        return

    def delete_client(self, client):
        """Delete the specified client
        Arguments:
          lock - either name, id or Client object
        Returns:
           None
        Raises:
          ValueError if lock not found
        """
        session = Session()
        if (type(client) is int):
            client = session.query(Client).filter_by(client_id=client).first()
            if client == None:
                raise ValueError("No client found for ID ")
        if (type(client) is str):
            client = session.query(Client).filter_by(name=client).first()
            if client == None:
                raise ValueError("No client found for name")
        try:
            logging.info("Deleting client " + str(client))
            self._lock_db(session)
            session.delete(client)
            session.commit()
            self._unlock_db(session)
        finally:
            session.close()
        return

    def get_resources(self):
        """Return all current resources
        Returns:
           List of Resource objects
        """
        session = Session()
        try:
            resources = session.query(Resource).all()
            for resource in resources:
                lazy_load(resource)
            return resources
        finally:
            session.close()
        return

    def delete_resource(self, resource):
        """Delete the specified client
        Arguments:
          lock - either uri, id or Resource object
        Returns:
           None
        Raises:
          ValueError if lock not found
        """
        session = Session()
        if (type(resource) is int):
            resource = session.query(Resource).filter_by(resource_id=resource).first()
            if resource == None:
                raise ValueError("No client found for ID ")
        if (type(resource) is str):
            resource = session.query(Resource).filter_by(uri=resource).first()
            if resource == None:
                raise ValueError("No client found for name")
        try:
            logging.info("Deleting resource " + str(resource))
            self._lock_db(session)
            session.delete(resource)
            session.commit()
            self._unlock_db(session)
        finally:
            session.close()
        return

    def _lock_db(self, session):
        """Utility to obtain a lock over the MySQL tables to ensure no race condition"""
        if (self.url.startswith('mysql')):
            session.execute('lock table client write, resource write, resource_lock write')

    def _unlock_db(self, session):
        """Utility to obtain release a lock over the MySQL tables"""
        if (self.url.startswith('mysql')):
            session.execute('unlock tables')



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import unittest
from tempfile import mkstemp

from ensembl.production.core.resource_lock import ResourceLocker, LockException

logging.basicConfig()

cname = 'testclient'
ruri = 'proto://myres/test'
rlock = 'read'
wlock = 'write'


def get_db():
    (handle, dbfile) = mkstemp(suffix='.db')
    os.close(handle)
    return dbfile


def run_tst(test_function):
    dbfile = get_db()
    try:
        locker = ResourceLocker('sqlite:///' + str(dbfile))
        test_function(locker)
        return
    finally:
        os.remove(dbfile)


class UtilsTest(unittest.TestCase):

    def test_client(self):
        def ctest(locker):
            client = locker.get_client(cname)
            self.assertEqual(cname, client.name, "Client name correct")
            self.assertTrue(client.client_id > 0, "Client has ID")
            client2 = locker.get_client(cname)
            self.assertEqual(client.client_id, client2.client_id, "Client retrieved")
            self.assertEqual(client.name, client2.name, "Client name retrieved")
            self.assertEqual(1, len(locker.get_clients()))
            locker.delete_client(cname)
            self.assertEqual(0, len(locker.get_clients()))

        run_tst(ctest)
        return

    def test_resource(self):
        def rtest(locker):
            res = locker.get_resource(ruri)
            self.assertEqual(ruri, res.uri, "Resource uri correct")
            self.assertTrue(res.resource_id > 0, "Resource has ID")
            res2 = locker.get_resource(ruri)
            self.assertEqual(res.resource_id, res2.resource_id, "Resource retrieved")
            self.assertEqual(res.uri, res2.uri, "Resource uri retrieved")
            self.assertEqual(1, len(locker.get_resources()))
            locker.delete_resource(ruri)
            self.assertEqual(0, len(locker.get_resources()))

        run_tst(rtest)
        return

    def test_read_lock(self):
        def rltest(locker):
            lock = locker.lock(cname, ruri, rlock)
            self.assertEqual(rlock, lock.lock_type, "Lock type correct")
            self.assertTrue(lock.resource_lock_id > 0, "Lock has ID")
            self.assertEqual(cname, lock.client.name, "Client correct")
            self.assertEqual(ruri, lock.resource.uri, "Resource correct")
            locks = [l for l in locker.get_locks() if (l.resource_lock_id == lock.resource_lock_id)]
            self.assertEqual(1, len(locks), "Lock exists")

            # test retrieval by client name
            self.assertEqual(1, len(locker.get_locks(client_name=cname)), "Lock found")
            self.assertEqual(0, len(locker.get_locks(client_name='badman')), "Lock not found")
            # test retrieval by resource
            self.assertEqual(1, len(locker.get_locks(resource_uri=ruri)), "Lock found")
            self.assertEqual(0, len(locker.get_locks(resource_uri='badres')), "Lock not found")
            # test retrieval by type  
            self.assertEqual(1, len(locker.get_locks(lock_type='read')), "Lock found")
            self.assertEqual(0, len(locker.get_locks(lock_type='write')), "Lock not found")
            # test retrieval by all
            self.assertEqual(1, len(locker.get_locks(lock_type='read', client_name=cname, resource_uri=ruri)),
                             "Lock found")
            self.assertEqual(0, len(locker.get_locks(lock_type='read', client_name='badclient', resource_uri=ruri)),
                             "Lock found")

            lock2 = locker.get_lock(lock.resource_lock_id)
            self.assertEqual(lock.resource_lock_id, lock2.resource_lock_id, "Lock ID correct")
            self.assertEqual(rlock, lock2.lock_type, "Lock type correct")
            self.assertEqual(cname, lock2.client.name, "Client correct")
            self.assertEqual(ruri, lock2.resource.uri, "Resource correct")
            locker.unlock(lock)
            locks = [l for l in locker.get_locks() if (l.resource_lock_id == lock.resource_lock_id)]
            self.assertEqual(0, len(locks), "Lock does not exist")
            lock3 = locker.get_lock(lock.resource_lock_id)
            self.assertEqual(None, lock3, "Lock not found")
            return

        run_tst(rltest)
        return

    def test_write_lock(self):
        def wltest(locker):
            lock = locker.lock(cname, ruri, wlock)
            self.assertEqual(wlock, lock.lock_type, "Lock type correct")
            self.assertTrue(lock.resource_lock_id > 0, "Lock has ID")
            self.assertEqual(cname, lock.client.name, "Client correct")
            self.assertEqual(ruri, lock.resource.uri, "Resource correct")
            locks = [l for l in locker.get_locks() if (l.resource_lock_id == lock.resource_lock_id)]
            self.assertEqual(1, len(locks), "Lock exists")

            # test retrieval by client name
            self.assertEqual(1, len(locker.get_locks(client_name=cname)), "Lock found")
            self.assertEqual(0, len(locker.get_locks(client_name='badman')), "Lock not found")
            # test retrieval by resource
            self.assertEqual(1, len(locker.get_locks(resource_uri=ruri)), "Lock found")
            self.assertEqual(0, len(locker.get_locks(resource_uri='badres')), "Lock not found")
            # test retrieval by type  
            self.assertEqual(1, len(locker.get_locks(lock_type='write')), "Lock found")
            self.assertEqual(0, len(locker.get_locks(lock_type='read')), "Lock not found")
            # test retrieval by all
            self.assertEqual(1, len(locker.get_locks(lock_type='write', client_name=cname, resource_uri=ruri)),
                             "Lock found")
            self.assertEqual(0, len(locker.get_locks(lock_type='write', client_name='badclient', resource_uri=ruri)),
                             "Lock found")

            locker.unlock(lock)
            locks = [l for l in locker.get_locks() if (l.resource_lock_id == lock.resource_lock_id)]
            self.assertEqual(0, len(locks), "Lock does not exist")
            return

        run_tst(wltest)
        return

    def test_readread_lock(self):
        def rltest(locker):
            lock1 = locker.lock(cname, ruri, rlock)
            self.assertEqual(rlock, lock1.lock_type, "Lock type correct")
            self.assertTrue(lock1.resource_lock_id > 0, "Lock has ID")
            self.assertEqual(cname, lock1.client.name, "Client correct")
            self.assertEqual(ruri, lock1.resource.uri, "Resource correct")

            lock2 = locker.lock(cname, ruri, rlock)
            self.assertEqual(rlock, lock2.lock_type, "Lock type correct")
            self.assertTrue(lock1.resource_lock_id > 0, "Lock has ID")
            self.assertEqual(cname, lock2.client.name, "Client correct")
            self.assertEqual(ruri, lock2.resource.uri, "Resource correct")
            return

        run_tst(rltest)
        return

    def test_writeread_lock(self):
        def wrtest(locker):
            lock1 = locker.lock(cname, ruri, wlock)
            self.assertEqual(wlock, lock1.lock_type, "Lock type correct")
            self.assertTrue(lock1.resource_lock_id > 0, "Lock has ID")
            self.assertEqual(cname, lock1.client.name, "Client correct")
            self.assertEqual(ruri, lock1.resource.uri, "Resource correct")
            with self.assertRaises(LockException):
                locker.lock(cname, ruri, rlock)
            # unlock and retry
            locker.unlock(lock1)
            locker.lock(cname, ruri, rlock)
            return

        run_tst(wrtest)
        return

    def test_readwrite_lock(self):
        def rwtest(locker):
            lock1 = locker.lock(cname, ruri, rlock)
            self.assertEqual(rlock, lock1.lock_type, "Lock type correct")
            self.assertTrue(lock1.resource_lock_id > 0, "Lock has ID")
            self.assertEqual(cname, lock1.client.name, "Client correct")
            self.assertEqual(ruri, lock1.resource.uri, "Resource correct")
            with self.assertRaises(LockException):
                locker.lock(cname, ruri, wlock)
                # unlock and retry
            locker.unlock(lock1)
            locker.lock(cname, ruri, wlock)
            return

        run_tst(rwtest)
        return



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Reporting module.
"""
import logging
import socket
from datetime import datetime


_HOSTNAME = socket.gethostname()


reporting_logger = logging.getLogger(__name__)


def make_report(report_type, msg, params=None, resource=''):
    """Return a dictionary compatible with ReportFormatter.
    """
    if params is None:
        params = {}
    return {
        'params': params,
        'resource': resource,
        'report_type': report_type,
        'msg': msg
    }


class ReportFormatter:
    """Formatter class creating messages for the report index.
    """
    def __init__(self, process_name):
        self.process_name = process_name

    def format(self, report):
        """Take a report templated dictionary and return a dictionary formatted
        for the report index.
        """
        try:
            report_type = report['report_type']
        except KeyError:
            raise ValueError("Report: %s is missing report required field 'report_type'" % report)
        return {
            'report_type': report_type, # CRITICAL, ERROR, INFO, DEBUG, WARN
            'host': _HOSTNAME,
            'process': self.process_name,
            'resource': report.get('resource', ''),
            'params': report.get('params', {}),
            'report_time': datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3],
            'message': report.get('msg', '')
        }




#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from sqlalchemy import create_engine, Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

Base = declarative_base()

__all__ = ['CoreInstance', 'get_division', 'get_release']


class Meta(Base):
    __tablename__ = 'meta'

    meta_id = Column(Integer, primary_key=True)
    meta_key = Column(String)
    meta_value = Column(String)

    def __repr__(self):
        return "<Meta(meta_id='%s',meta_key='%s', meta_value='%s')>" % (
            self.meta_id, self.meta_key, self.meta_value)


Session = sessionmaker()


class CoreInstance:
    __division = None
    __release = None

    def __init__(self, url, timeout=3600):
        self.engine = create_engine(url, pool_recycle=timeout, echo=False)
        Session.configure(bind=self.engine)

    def __get_meta_value(self, meta_key):
        s = Session()
        try:
            meta = s.query(Meta).filter(Meta.meta_key == meta_key).first()
            return meta.meta_value
        finally:
            s.close()

    @property
    def division(self):
        if self.__division is None:
            self.__division = self.__get_meta_value('species.division')
        return self.__division

    @property
    def release(self):
        if self.__release is None:
            self.__release = self.__get_meta_value('schema_version')
        return self.__release


def get_division(src_uri, tgt_uri, db_type):
    uri = src_uri
    if db_type == "variation" or db_type == "funcgen":
        uri = tgt_uri.replace("_variation_", "_core_").replace("_funcgen_", "_core_")
    inst = CoreInstance(uri)
    division_meta = inst.division
    division = str(division_meta).replace('Ensembl', '')
    return division.lower()


def get_release(src_uri):
    inst = CoreInstance(src_uri)
    return int(inst.release)



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from contextlib import contextmanager
import logging
from kombu import Connection, Exchange
from kombu.pools import producers


OPTIONS = {
    'serializer': 'json',
    'delivery_mode': 2,
    'retry': True,
    # 'compression': 'zlib',
    'retry_policy': {
        'interval_start': 0,
        'interval_step': 2,
        'interval_max': 30,
        'max_retries': 5
    }
}


logger = logging.getLogger(__name__)


class AMQPPublisher:
    """Simple publisher for connecting and publishing messages to a AMQP broker exchange.
    An optional formatter instance can be passed in order to format the message before sending it,
    (it must implement a method named `format`).
    Connections are lazy and stored in a global pool so that AMQPPublisher instances with the same uri
    in the same process will use the same connection object.
    The additional kwargs `options` are passed as additional parameters to kombu.Producer.publish
    please refer to: https://kombu.readthedocs.io/en/latest/userguide/producers.html#reference
    """
    def __init__(self, uri, exchange_name, exchange_type='topic', routing_key=None, formatter=None, **options):
        self.routing_key = routing_key
        self.formatter = formatter
        self.options = {**OPTIONS, **options}
        self.connection = Connection(uri, transport_options=self.options.get('retry_policy'))
        self.exchange = Exchange(exchange_name, type=exchange_type)


    class AMQPProducer:
        def __init__(self, producer, exchange, routing_key, formatter, options):
            self.producer = producer
            self.exchange = exchange
            self.routing_key = routing_key
            self.formatter = formatter
            self.options = options

        def publish(self, msg, routing_key=None):
            body = self.formatter.format(msg) if self.formatter is not None else msg
            key = routing_key if routing_key else self.routing_key
            if not key:
                raise ValueError('Invalid routing_key: {} (Producer routing_key is: {})'.format(
                    routing_key, self.routing_key
                    ))
            logger.debug(
                "Publishing message: %s, %s, Routing Key: %s, options: %s", body, self.exchange, key, self.options
            )
            self.producer.publish(body,
                                  exchange=self.exchange,
                                  routing_key=key,
                                  declare=(self.exchange,),
                                  **self.options)
            logger.debug(
                'Published AMQP message: %s, %s, Routing Key: %s', body, self.exchange, key
            )


    @contextmanager
    def acquire_producer(self, block=True):
        """Acquire a producer from the global pool for holding a connection/channel while publishing.
        Useful for removing the overhead of acquiring a producer from the global pool when
        publishing several messages in a row.
        """
        with producers[self.connection].acquire(block=block) as producer:
            logger.debug('Acquired producer for connection %s', self.connection.as_uri())
            yield self.AMQPProducer(producer,
                                    self.exchange,
                                    self.routing_key,
                                    self.formatter,
                                    self.options)
        logger.debug('Released producer for connection %s', self.connection.as_uri())

    def publish(self, msg, routing_key=None):
        """Acquire a producer from the global pool and publish a single message.
        An optional `routing_key` can be passed. If no `routing_key` is passed
        the instance one is used (in that case, if no `routing_key` has been set in the instance)
        a ValueError is raised. Also raises kombu.exceptions.OperationalError if the message
        cannot be delivered accordig to policy specified in this instance (i.e in self.options)
        """
        with self.acquire_producer() as producer:
            producer.publish(msg, routing_key)




#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import logging
import unittest
import os
import sys

from ensembl.production.core.config import parse_debug_var
import ensembl.production.core.server_utils as su

LOG_LEVEL = logging.DEBUG if parse_debug_var(os.getenv("DEBUG")) else logging.WARNING

logging.basicConfig(
    stream=sys.stdout,
    format="%(asctime)s %(levelname)-8s %(name)s: %(message)s",
    level=LOG_LEVEL,
)
logger = logging.getLogger(__name__)

class ServerTest(unittest.TestCase):
    def test_status(self):
        status = su.get_status()
        logger.debug(status)
        self.assertTrue(status["n_cpus"] >= 1)
        self.assertTrue(status["load_1m"] >= 0)
        self.assertTrue(status["load_5m"] >= 0)
        self.assertTrue(status["load_15m"] >= 0)
        self.assertTrue(status["memory_total_m"] >= 0)
        self.assertTrue(status["memory_used_m"] >= 0)
        self.assertTrue(status["memory_available_m"] >= 0)
        self.assertTrue(status["memory_used_pct"] >= 0)
        self.assertTrue(status["memory_used_pct"] <= 100)

    def test_status_dir(self):
        status = su.get_status(dir_name="/")
        logger.debug(status)
        self.assertTrue(status["n_cpus"] >= 1)
        self.assertTrue(status["load_1m"] >= 0)
        self.assertTrue(status["load_5m"] >= 0)
        self.assertTrue(status["load_15m"] >= 0)
        self.assertTrue(status["memory_total_m"] >= 0)
        self.assertTrue(status["memory_used_m"] >= 0)
        self.assertTrue(status["memory_available_m"] >= 0)
        self.assertTrue(status["memory_used_pct"] >= 0)
        self.assertTrue(status["memory_used_pct"] <= 100)
        self.assertTrue(status["disk_total_g"] >= 0)
        self.assertTrue(status["disk_used_g"] >= 0)
        self.assertTrue(status["disk_available_g"] >= 0)
        self.assertTrue(status["disk_used_pct"] >= 0)
        self.assertTrue(status["disk_used_pct"] <= 100)


class AssertTest(unittest.TestCase):
    def test_raises_assert_http_uri(self):
        self.assertRaises(ValueError, su.assert_http_uri, '')
        self.assertRaises(ValueError, su.assert_http_uri, 'invalid_uri')
        self.assertRaises(ValueError, su.assert_http_uri, 'http://uri-with-no-slash')
        self.assertRaises(ValueError, su.assert_http_uri, 'mysql://wrong-schema')

    def test_passes_assert_http_uri(self):
        su.assert_http_uri('http://server-name/')
        su.assert_http_uri('https://server-name:port/')

    def test_raises_assert_mysql_uri(self):
        self.assertRaises(ValueError, su.assert_mysql_uri, '')
        self.assertRaises(ValueError, su.assert_mysql_uri, 'invalid_uri')
        self.assertRaises(ValueError, su.assert_mysql_uri, 'http://wrong_schema')
        self.assertRaises(ValueError, su.assert_mysql_uri, 'mysql://invalid')
        self.assertRaises(ValueError, su.assert_mysql_uri, 'mysql://user@server-no-slash')
        self.assertRaises(ValueError, su.assert_mysql_uri, 'mysql://user:pass@server-no-port/')

    def test_passes_assert_mysql_uri(self):
        su.assert_mysql_uri('mysql://user@server:3006/')
        su.assert_mysql_uri('mysql://user:pass@server:3306/')

    def test_raises_assert_mysql_db_uri(self):
        self.assertRaises(ValueError, su.assert_mysql_db_uri, '')
        self.assertRaises(ValueError, su.assert_mysql_db_uri, 'invalid_uri')
        self.assertRaises(ValueError, su.assert_mysql_db_uri, 'http://wrong_schema')
        self.assertRaises(ValueError, su.assert_mysql_db_uri, 'mysql://invalid')
        self.assertRaises(ValueError, su.assert_mysql_db_uri, 'mysql://user:pass@server-no-slash')
        self.assertRaises(ValueError, su.assert_mysql_db_uri, 'mysql://user:pass@server-no-db:3306/')

    def test_pass_assert_mysql_db_uri(self):
        su.assert_mysql_db_uri('mysql://user:pass@server:3306/db_name')

    def test_raises_assert_email(self):
        self.assertRaises(ValueError, su.assert_email, '')
        self.assertRaises(ValueError, su.assert_email, 'invalid_email.com')

    def test_passes_assert_email(self):
        su.assert_email('valid.email@domain.ac.uk')



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import re


def dict_to_perl_string(input_dict):
    """Transform the supplied dict into a string representation of a Perl hash"""
    pairs = []
    for k, v in sorted(filter(lambda k_v: k_v[1] != None, input_dict.items())):
        k = str(k)
        t = type(v).__name__
        if t == 'str':
            pairs.append("\"%s\" => \"%s\"" % (k, escape_perl_string(v)))
        elif t == 'int':
            pairs.append("\"%s\" => %d" % (k, v))
        elif t == 'float':
            pairs.append("\"%s\" => %f" % (k, v))
        elif t == 'list':
            pairs.append("\"%s\" => %s" % (k, list_to_perl_string(v)))
        elif t == 'dict':
            pairs.append("\"%s\" => %s" % (k, dict_to_perl_string(v)))
        elif t == 'bool':
            if str(v) == "True":
                pairs.append("\"%s\" => %d" % (k, 1))
        else:
            raise Exception("Unsupported type " + str(t))
    return "{%s}" % ", ".join(pairs)


def list_to_perl_string(input_list):
    """Transform the supplied array into a string representation of a Perl array"""
    elems = []
    for v in input_list:
        t = type(v).__name__
        if t == 'str':
            elems.append("\"%s\"" % escape_perl_string(v))
        elif t == 'int':
            elems.append("%d" % v)
        elif t == 'float':
            elems.append("%f" % v)
        elif t == 'list':
            elems.append("%s" % list_to_perl_string(v))
        elif t == 'dict':
            elems.append("%s" % dict_to_perl_string(v))
        else:
            raise Exception("Unsupported type " + str(t))
    return "[%s]" % ", ".join(elems)


def escape_perl_string(v):
    """Escape characters with special meaning in perl"""
    s = str(v).replace("$", "\\$").replace("\"", "\\\"").replace("@", "\\@")
    return re.sub('[\n\t\r]+', ' ', s)


def perl_string_to_python(s):
    """Parse a Perl hash string into a Python dict"""
    s = s.replace("=>", ":").replace("\\$", "$").replace("\\@", "@")
    s = re.sub('[\n\t\r]+', ' ', s)
    try:
        res = json.loads(s)
    except json.JSONDecodeError as e:
        beg = max(0, e.pos - 25)
        end = min(len(e.doc), e.pos + 25)
        raise ValueError('Invalid JSON: {}. --> {} <--'.format(e, e.doc[beg:end]))
    return res



#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ensembl.production.core.rest import RestClient
from ensembl.production.core.utils import json_decode_error_context


class GenomeMetadataRestClient(RestClient):
    """
       Client for interacting with the Ensembl metadata REST API.
    """

    dataset_endpoint = '{}api/genome_metadata/datasets/'
    dataset_uuid_endpoint = '{}api/genome_metadata/datasets/{}'
    genome_endpoint = '{}api/genome_metadata/genomes/'
    genome_uuid_endpoint = '{}api/genome_metadata/genomes/{}'

    def __init__(self, uri):
        super().__init__(uri)

    def _session(self, use_ssl=False):
        session = super()._session(use_ssl)
        return session

    def create_dataset(self, dataset):
        """
        Create a new dataset using the provided details.
        """
        with self._session() as session:
            r = session.post(self.dataset_endpoint.format(self.uri), json=dataset)
        r.raise_for_status()
        return r.json()

    def get_all_datasets(self):
        """
        Get all datasets.
        """
        with self._session() as session:
            r = session.get(self.dataset_endpoint.format(self.uri))
        r.raise_for_status()
        return r.json()

    def get_dataset_by_uuid(self, uuid):
        """
        Get a dataset by its UUID.
        """
        with self._session() as session:
            r = session.get(self.dataset_uuid_endpoint.format(self.uri, uuid))
        r.raise_for_status()
        return r.json()

    def get_all_genomes(self):
        """
        Get all genomes.
        """
        with self._session() as session:
            r = session.get(self.genome_endpoint.format(self.uri))
        r.raise_for_status()
        return r.json()

    def get_genome_by_uuid(self, uuid):
        """
        Get a genome by its UUID.
        """
        with self._session() as session:
            r = session.get(self.genome_uuid_endpoint.format(self.uri, uuid))
        r.raise_for_status()
        return r.json()




#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import re
from datetime import datetime

import requests
from sqlalchemy.engine import make_url

from ensembl.production.core.server_utils import assert_http_uri, assert_mysql_db_uri, assert_email


class HandoverClient(object):
    """
    Client for submitting databases for handover
    """

    handovers = '{}jobs'
    handover_token = '{}jobs/{}'

    def __init__(self, uri):
        assert_http_uri(uri)
        self.uri = uri

    def submit_handover(self, spec):
        """
        Arguments:
          spec : dict containing keys `src_uri`, `comment` and `contact`
        #TODO move this onto submit_job standard from parent class
        """
        try:
            assert_mysql_db_uri(spec['src_uri'])
            assert_email(spec['contact'])
            logging.info("Submitting {} for handover".format(spec['src_uri']))
            logging.debug(spec)
            r = requests.post(self.handovers.format(self.uri), json=spec)
            r.raise_for_status()
            return r.json()
        except Exception as e : 
            error_msg = r.json()
            raise RuntimeError(error_msg)

    def list_handovers(self):
        """
        Retrieve full list of handover databases
        """
        logging.info("Listing from %s", self.handovers.format(self.uri))
        r = requests.get(self.handovers.format(self.uri))
        r.raise_for_status()
        return r.json()

    def print_handover_detail(self, handover):
        """
        Print out details of a handover
        Arguments:
          handover : Handover dict
        """
        handover_api_link = self.handover_token.format(self.uri, str(handover['handover_token']))
        report_time = datetime.strptime(handover['report_time'], "%Y-%m-%dT%H:%M:%S.%f")
        if 'current_message' in handover:
            logging.info("Handover %s (%s) submitted by (%s) - %s on %s" % (
                handover_api_link, handover['src_uri'], handover['contact'], handover['current_message'],
                report_time.strftime('%d-%m-%Y %H:%M')))
        elif 'message' in handover:
            logging.info("Handover %s (%s) submitted by (%s) - %s on %s." % (
                handover_api_link, handover['src_uri'], handover['contact'], handover['message'],
                report_time.strftime('%d-%m-%Y %H:%M'), ))

    def retrieve_handover(self, handover_token):
        """
        Retrieve a handover using an handover_token
        Arguments:
          handover_token: handover token, e.g: 56bf1f7e-ebdf-11e8-8afa-005056ab4d6f
        """
        logging.info("Retrieving details for handover " + str(handover_token))
        r = requests.get(self.handover_token.format(self.uri, str(handover_token)))
        r.raise_for_status()
        return r.json()

    def handover_summary_email(self, handovers, email):
        """
        Retrieve all the handovers associated with a given email
        Generate a unique list of handed over databases
        If a database was handed over multiple times, the latest one will be displayed.
        Print everything
        """
        fail_pattern = re.compile(".*(failed|problems).*")
        successful_pattern = re.compile(".*successful.*")
        summary = {}
        logging.info("Retrieving handovers for %s", str(email) )
        for handover in handovers:
            if handover['contact'] == email:
                src_uri = make_url(handover['src_uri'])
                if src_uri.database not in summary:
                    summary[src_uri.database] = handover
        for sum in summary:
            handover_result = "in progress"
            if fail_pattern.match(summary[sum]['current_message']):
                handover_result = "failed"
            elif successful_pattern.match(summary[sum]['current_message']):
                handover_result = "success"
            logging.info("Handover %s - %s : %s" % (summary[sum]['handover_token'], sum, handover_result))



#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest
from ensembl.production.core.db_introspects import _apply_filters


class DBIntrospectsTest(unittest.TestCase):
    def test_apply_filters_no_filters(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        result = _apply_filters(names_list, incl_filters=None, skip_filters=None)
        self.assertSetEqual(set(names_list), result)

    def test_apply_filters_incl_filters(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        incl_filters = ['.*core_105.*']
        expected = [
            'species1_core_105_1',
            'species2_core_105_1',
        ]
        result = _apply_filters(names_list, incl_filters=incl_filters, skip_filters=None)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_incl_filters_no_wildcards(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        incl_filters = ['core_105']
        expected = []
        result = _apply_filters(names_list, incl_filters=incl_filters, skip_filters=None)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_incl_filters_and_matches(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        incl_filters = ['.*core_105.*', 'species4_core_104_2']
        expected = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species4_core_104_2',
        ]
        result = _apply_filters(names_list, incl_filters=incl_filters, skip_filters=None)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_incl_filters_and_matches_skip_filters(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        incl_filters = ['.*core_105.*', 'species4_core_104_2']
        skip_filters = ['species2.*']
        expected = [
            'species1_core_105_1',
            'species4_core_104_2',
        ]
        result = _apply_filters(names_list, incl_filters=incl_filters, skip_filters=skip_filters)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_incl_filters_and_matches_skip_filters_no_wildcards(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        incl_filters = ['.*core_105.*', 'species4_core_104_2']
        skip_filters = ['species2']
        expected = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species4_core_104_2',
        ]
        result = _apply_filters(names_list, incl_filters=incl_filters, skip_filters=skip_filters)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_incl_filters_and_matches_skip_filters_and_matches(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        incl_filters = ['.*core_105.*', 'species4_core_104_2']
        skip_filters = ['species2.*', 'species1_core_105_1']
        expected = [
            'species4_core_104_2',
        ]
        result = _apply_filters(names_list, incl_filters=incl_filters, skip_filters=skip_filters)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_skip_filters(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        skip_filters = ['.*core_104.*']
        expected = [
            'species1_core_105_1',
            'species2_core_105_1',
        ]
        result = _apply_filters(names_list, incl_filters=None, skip_filters=skip_filters)
        self.assertSetEqual(set(expected), result)

    def test_apply_filters_skip_filters_and_matches(self):
        names_list = [
            'species1_core_105_1',
            'species2_core_105_1',
            'species3_core_104_2',
            'species4_core_104_2',
        ]
        skip_filters = ['.*core_104.*', 'species1_core_105_1']
        expected = [
            'species2_core_105_1',
        ]
        result = _apply_filters(names_list, incl_filters=None, skip_filters=skip_filters)
        self.assertSetEqual(set(expected), result)




#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from json import JSONDecodeError
import os
from urllib.parse import urlsplit, urlunsplit
from requests import RequestException
from ensembl.production.core.rest import RestClient
from ensembl.production.core.utils import json_decode_error_context


class DbCopyRestClient(RestClient):
    """
    Client for submitting database copy jobs to the db copy REST API
    """

    jobs = '{}'
    jobs_id = '{}/{}'
    src_host_list_url = 'srchost'
    tgt_host_list_url = 'tgthost'

    def submit_job(self, src_host, src_incl_db, src_skip_db, src_incl_tables,
                   src_skip_tables, tgt_host, tgt_db_name, skip_optimize,
                   wipe_target, convert_innodb, email_list, user):
        """
        Submit a new job
        Arguments:
          src_host : Source host for the copy (host:port)
          src_incl_db : List of database to include in the copy. If not defined all the databases from the server will be copied
          src_skip_db : List of database to exclude from the copy.
          src_incl_tables : List of tables to include in the copy.
          src_skip_tables : List of tables to exclude from the copy.
          tgt_host : List of hosts to copy to (host:port,host:port)
          tgt_db_name : Name of database on target server. Used for renaming databases
          skip_optimize : Skip the database optimization step after the copy. Useful for very large databases
          wipe_target: Delete database on target before copy
          convert_innodb: Convert innoDB tables to MyISAM
          email_list: List of emails
          user: user name
        """
        #FIXME Make signature identical to parent
        payload = {
            'src_host': src_host,
            'src_incl_db': src_incl_db,
            'src_skip_db': src_skip_db,
            'src_incl_tables': src_incl_tables,
            'src_skip_tables': src_skip_tables,
            'tgt_host': tgt_host,
            'tgt_db_name': tgt_db_name,
            'skip_optimize': skip_optimize,
            'wipe_target': wipe_target,
            'convert_innodb': convert_innodb,
            'email_list': email_list,
            'user': user,
        }
        try:
            return super().submit_job(payload)
        except RequestException as err:
            raise RuntimeError(str(err)) from err
        except KeyError as err:
            raise RuntimeError(f"API response error. Missing field: '{err}'") from err

    def retrieve_job(self, job_id):
        """
        Retrieve information on a job.
        Arguments:
          job_id - ID of job to retrieve

        Raises:
          RuntimeError: If the request for jobs fails
        """
        try:
            return super().retrieve_job(job_id)
        except RequestException as err:
            raise RuntimeError(str(err)) from err

    def list_jobs(self):
        """
        List all current jobs
        Raises:
          RuntimeError: If the request for jobs fails
        """
        try:
            return super().list_jobs()
        except RequestException as err:
            raise RuntimeError(str(err)) from err

    def print_job(self, job, user, print_results=False):
        """
        Print out details of a job
        Arguments:
          job : Job to render
          print_results : set to True to show results
          user: name of the user to filter on
        """
        if 'url' in job:
            if user:
                if user == job['user']:
                    logging.info("Job %s from (%s) to (%s) by %s - status: %s",
                                 job['url'], job['src_host'], job['tgt_host'], job['user'], job['overall_status'])
            else:
                logging.info("Job %s from (%s) to (%s) by %s - status: %s",
                             job['url'], job['src_host'], job['tgt_host'], job['user'], job['overall_status'])
        else:
            if user:
                if user == job['user']:
                    logging.info("Job %s from (%s) to (%s) by %s - status: %s",
                                 job['job_id'], job['src_host'], job['tgt_host'], job['user'], job['overall_status'])
            else:
                logging.info("Job %s from (%s) to (%s) by %s - status: %s",
                             job['job_id'], job['src_host'], job['tgt_host'], job['user'], job['overall_status'])
        if job['overall_status'] == 'Running':
            if print_results == True:
                logging.info("Copy status: %s", job['overall_status'])
                logging.info("%s complete", job['detailed_status']['progress'])

    def print_inputs(self, i):
        """
        Print out details of job input
        Arguments:
          i : job input
        """
        logging.info("Source host: %s", i['src_host'])
        logging.info("Target hosts: %s", i['tgt_host'])
        logging.info("Detailed parameters:")
        logging.info("%s", i)

    def check_hosts(self, host_type, urls):
        hosts = self.retrieve_host_list(host_type)['results']
        # Create dict with valid hostnames as keys and respective valid ports as values
        host_port_map = dict(list(map(lambda x: (x['name'], x['port']), hosts)))
        errs = []
        for url in urls:
            err = self._check_host(url, host_port_map)
            if err:
                errs.append(err)
        return errs

    def _check_host(self, url, host_port_map):
        host, port = url.split(':')
        host_parts = host.split('.')
        if len(host_parts) > 1:
            if host.endswith('.ebi.ac.uk'):
                return 'Invalid domain: {}'.format(host)
        hostname = host_parts[0]
        actual_port = host_port_map.get(hostname)
        if actual_port is None:
            return 'Invalid hostname: {}'.format(host)
        if int(port) != int(actual_port):
            return 'Invalid port for hostname: {}. Please use port: {}'.format(host, actual_port)

    def retrieve_host_list(self, host_type):
        if host_type == 'source':
            url = self.src_host_list_url
        elif host_type == 'target':
            url = self.tgt_host_list_url
        else:
            raise ValueError('Invalid host_type: %s. Use "source" or "target"' % host_type)
        # Deconstruct the url in order to work with new endpoints.
        # TODO: Refactor when the old db_copy service is retired.
        uri = urlsplit(self.uri)
        endpoint = urlunsplit((uri.scheme,
                               uri.netloc,
                               os.path.join(os.path.dirname(uri.path), url),
                               uri.query,
                               uri.fragment))
        with self._session() as session:
            r = session.get(endpoint)
        if r.status_code != 200:
            logging.error("Failed to retrieve host list: %s", r.text)
        try:
            r.raise_for_status()
        except RequestException as err:
            raise RuntimeError(str(err)) from err
        try:
            return r.json()
        except JSONDecodeError as err:
            raise RuntimeError(f"Can't decode JSON response: {json_decode_error_context(err)}")



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import unittest
import pathlib
import sys
from shutil import copy2

from ensembl.production.core.config import parse_debug_var
from ensembl.production.core.models.hive import HiveInstance


here = pathlib.Path(__file__).parent.resolve()

LOG_LEVEL = logging.DEBUG if parse_debug_var(os.getenv("DEBUG")) else logging.WARNING

logging.basicConfig(
    stream=sys.stdout,
    format="%(asctime)s %(levelname)-8s %(name)s: %(message)s",
    level=LOG_LEVEL,
)

logger = logging.getLogger(__name__)
logging.getLogger('sqlalchemy.engine').setLevel(LOG_LEVEL)

DB_TEMPLATE_FILENAME = "test_pipeline.db.template"
DB_FILENAME = "test_pipeline.db.sqlite3"


class HiveTest(unittest.TestCase):
    """Create fresh database file"""

    def setUp(self):
        logger.info("Creating test sqlite database")
        copy2(here/DB_TEMPLATE_FILENAME, here/DB_FILENAME)
        logger.info("Connecting to hive test sqlite database %s", here/DB_FILENAME)
        self.hive = HiveInstance(f"sqlite:///{here/DB_FILENAME}")

    def tearDown(self):
        """Remove test database file"""
        logger.info("Removing test sqlite database")
        os.remove(here/DB_FILENAME)

    def test_create_job(self):
        """Basic test case for creating a new job"""
        job1 = self.hive.create_job('TestRunnable', {'x': 'y', 'a': 'b'})
        logger.debug(job1)
        job2 = self.hive.get_job_by_id(job1.job_id)
        logger.debug(job2)
        self.assertEqual(job1.job_id, job2.job_id)
        self.assertEqual(job1.analysis.logic_name, job2.analysis.logic_name)
        self.assertEqual(job1.input_id, job2.input_id)

    def test_check_semaphore_success(self):
        """Test case for checking on a finished semaphore"""
        semaphore_data = self.hive.get_semaphore_data(2)
        logger.debug(semaphore_data)
        status = self.hive.check_semaphores_for_job(semaphore_data)
        logger.debug("Status for 2 is " + status)
        self.assertEqual(status, 'complete', "Checking expected status for completed semaphore")

    def test_check_semaphore_incomplete(self):
        """Test case for checking on an incomplete semaphore"""
        semaphore_data = self.hive.get_semaphore_data(8)
        logger.debug("semaphore_data: %s", semaphore_data)
        status = self.hive.check_semaphores_for_job(semaphore_data)
        logger.debug("Status for 8 is " + status)
        self.assertEqual(status, 'incomplete', "Checking expected status for failed semaphore")

    def test_check_job_success(self):
        """Test case for checking on a completed single job"""
        job = self.hive.get_job_by_id(20)
        logger.debug(job)
        status = self.hive.get_job_tree_status(job)
        self.assertEqual("complete", status, "Checking status of completed single job")

    def test_check_job_failure(self):
        """Test case for checking on a failed single job"""
        job = self.hive.get_job_by_id(11)
        logger.debug(job)
        status = self.hive.get_job_tree_status(job)
        self.assertEqual("failed", status, "Checking status of failed single job")

    def test_check_job_tree_success(self):
        """Test case for checking on a completed job factory"""
        job = self.hive.get_job_by_id(1)
        logger.debug(job)
        status = self.hive.get_job_tree_status(job)
        logger.debug(status)
        self.assertEqual("complete", status, "Checking status of completed job factory")

    def test_check_job_tree_incomplete(self):
        """Test case for checking on an incomplete job factory"""
        job = self.hive.get_job_by_id(7)
        logger.debug("Job: %s", job)
        status = self.hive.get_job_tree_status(job)
        self.assertEqual("incomplete", status, "Checking status of incomplete job factory")

    def test_get_job_output_success(self):
        """Test case for getting output on a completed job factory"""
        output = self.hive.get_result_for_job_id(1)
        logger.debug(output)
        self.assertEqual('complete', output['status'], "Checking status of successful job factory output")
        self.assertTrue(output['output'] != None, "Checking output of successful job factory output")

    def test_get_job_output_incomplete(self):
        """Test case for getting output an incomplete job factory"""
        output = self.hive.get_result_for_job_id(7)
        logger.debug(output)
        self.assertEqual('incomplete', output['status'], "Checking status of incomplete job factory output")
        self.assertTrue('output' not in output, "Checking output of incomplete job factory output")

    def test_get_all_results(self):
        """Test case for listing all jobs"""
        jobs = self.hive.get_all_results('TestRunnable')
        self.assertEqual(1, len(jobs), "Checking we got just one job")

    def test_delete_job(self):
        job = self.hive.create_job('TestRunnable', {'x': 'y', 'a': 'b'})
        job_id = self.hive.get_job_by_id(job.job_id).job_id
        self.assertEqual(job.job_id, job_id)
        self.hive.delete_job(job)
        self.assertRaises(ValueError, self.hive.get_job_by_id, job.job_id)


if __name__ == '__main__':
    unittest.main()



#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

import requests

from ensembl.production.core.rest import RestClient


class EventClient(RestClient):
    """
    Simple client for submitting an event to the event service and checking on progress
    This uses the base RestClient, but all endpoint URIs for checking on submited events
    have process as a path element, so this client combines the job_id and process together
    """
    def __init__(self, uri):
        super(EventClient, self).__init__(uri)

    def submit_job(self, payload):
        """
        Submit a job using the supplied dict as payload. No checking is carried out on the payload
        Arguments:
          payload : job input as dict
        """
        logging.info("Submitting job")
        logging.debug(payload)
        with self._session() as session:
            r = session.post(f'{self.uri}/submit' , json=payload)
        if r.status_code != 201:
            logging.error("failed to submit workflow because: %s", r.text)
        r.raise_for_status()
        return r.json()

    def list_workflows(self, handover_token=None):
        """List all Workflows"""
        logging.info("Listing")
        url = f'{self.uri}/{handover_token}' if handover_token else self.uri
        r = requests.get(self.uri)
        r.raise_for_status()
        return r.json()

    def stop_workflow(self, payload):
        """
        Stop Running Workflow
        Arguments:
          payload : {
                        "handover_token": "3729-jhshs-12929-1mssn",
                        "job_id": "string",
                        "pipeline_name": "string"
                    }
        """        
        logging.info("Stop Workflow")
        logging.debug(payload)
        with self._session() as session:
            r = session.post(f'{self.uri}/stop', json=payload)
        if r.status_code != 201:
            logging.error("failed to stop workflow because: %s", r.text)
        r.raise_for_status()
        return r.json()

    def restart_workflow(self, payload):
        """
        Restart Stopped Workflow
        Arguments:
          payload : {
                        "handover_token": "3729-jhshs-12929-1mssn",
                        "restart_type": "BEEKEEPER"
                    }
        """   
        logging.info("Stop Workflow")
        logging.debug(payload)
        with self._session() as session:
            r = session.post(f'{self.uri}/restart', json=payload)
        if r.status_code != 201:
            logging.error("failed to restart workflow because: %s", r.text)
        r.raise_for_status()
        return r.json() 





#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from functools import lru_cache
import sqlalchemy as sa
import re


@lru_cache(maxsize=None)
def get_engine(hostname, port='3306', user='ensro', password='', database=''):
    uri = f'mysql://{user}:{password}@{hostname}:{port}'
    if database != '':
        uri += f'/{database}'
    return sa.create_engine(uri, pool_recycle=3600)


def get_schema_names(engine):
    return sa.inspect(engine).get_schema_names()


def get_table_names(engine, database):
    try:
        return sa.inspect(engine).get_table_names(schema=database)
    except sa.exc.OperationalError as e:
        raise ValueError('Invalid database: {}'.format(database)) from e


def _filter_names(names_list, filters):
    filter_names = set()
    filter_regexes = []
    if filters:
        try:
            for filter_name in filters:
                filter_regex = '^{}$'.format(filter_name)
                re.compile(filter_regex)
                filter_regexes.append(filter_regex)
        except re.error as e:
            raise ValueError('Invalid name_filter: {}'.format(filter_name)) from e
        filter_names_re = re.compile('|'.join(filter_regexes))
        filter_names = set(filter(filter_names_re.match, names_list))
    return filter_names


def _apply_filters(names_list, incl_filters, skip_filters):
    names = set(names_list)
    if incl_filters:
        names = _filter_names(names_list, incl_filters)
    skip_names = _filter_names(names_list, skip_filters)
    return names.difference(skip_names)


def get_database_set(hostname, port, user='ensro', password='', incl_filters=None, skip_filters=None):
    try:
        db_engine = get_engine(hostname, port, user, password)
    except RuntimeError as e:
        raise ValueError('Invalid hostname: {} or port: {}'.format(hostname, port)) from e
    database_list = get_schema_names(db_engine)
    return _apply_filters(database_list, incl_filters, skip_filters)


def get_table_set(hostname, port, database, user='ensro', password='', incl_filters=None, skip_filters=None):
    try:
        db_engine = get_engine(hostname, port, user, password)
    except RuntimeError as e:
        raise ValueError('Invalid hostname: {} or port: {}'.format(hostname, port)) from e
    table_list = get_table_names(db_engine, database)
    return _apply_filters(table_list, incl_filters, skip_filters)


def get_table_engines(hostname, port, database, user='ensro', password=''):
    try:
        db_engine = get_engine(hostname, port, user, password, database)
        with db_engine.connect() as connection:
            rp = connection.exec_driver_sql("SHOW TABLE STATUS")
            # table_name => table_engine
            return {r[0]: r[1] for r in rp}
    except RuntimeError as e:
        raise ValueError('Invalid hostname: {} or port: {}'.format(hostname, port)) from e



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import unittest
import os
import pathlib

from ensembl.production.core.config import load_config_yaml

logging.basicConfig()

here = pathlib.Path(__file__).parent.resolve()

class YamlLoaderTest(unittest.TestCase):
    def test_parse_simple_str(self):
        home_dir = os.getenv("HOME")
        os.environ['EXPECTEDVAR'] = 'environmentVar'
        logging.info("Expected Home is %s" % home_dir)
        config = load_config_yaml(here/'sample_config.yaml')
        self.assertIn(home_dir, config['base_dir'], "Home dir has been replaced")
        self.assertIn('environmentVar', config['server_names_file'], "Env var has been replaced in line")
        self.assertIn(home_dir, config['server_names_file'], "Home dir has been replaced in other line")



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
import pathlib
from setuptools import setup, find_namespace_packages

here = pathlib.Path(__file__).parent.resolve()
readme = (here / 'README.md').read_text(encoding='utf-8')
version = (here / 'VERSION').read_text(encoding='utf-8').strip()


def import_requirements():
    """Import ``requirements.txt`` file located at the root of the repository."""
    with (here / 'requirements.txt').open(encoding='utf-8') as file:
        return [line.strip() for line in file.readlines()]


setup(
    name='ensembl-prodinf-core',
    description='Ensembl Production infrastructure core package',
    long_description=readme,
    version=os.getenv('CI_COMMIT_TAG', version),
    namespace_packages=['ensembl'],
    packages=find_namespace_packages(where='src', include=['ensembl*'], exclude=['test']),
    package_dir={'': 'src'},
    include_package_data=True,
    url='https://github.com/Ensembl/ensembl-prodinf-core',
    license='Apache 2.0',
    author='Marc Chakiachvili,James Allen,Luca Da Rin Fioretto,Vinay Kaikala',
    author_email='mchakiachvili@ebi.ac.uk,jallen@ebi.ac.uk,ldrf@ebi.ac.uk,vkaikala@ebi.ac.uk',
    maintainer='Ensembl Production Team',
    maintainer_email='ensembl-production@ebi.ac.uk',
    python_requires='>=3.10',
    install_requires=import_requirements(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Utilities",
        "Topic :: System :: Distributed Computing",
    ]
)



#   See the NOTICE file distributed with this work for additional information
#   regarding copyright ownership.
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import re

from ensembl.production.core.rest import RestClient
from ensembl.production.core.server_utils import assert_mysql_db_uri


class MetadataClient(RestClient):

    def submit_job(self, database_uri, e_release, eg_release, release_date, current_release, email, comment, source):
        #FIXME Make signature identical to parent
        assert_mysql_db_uri(database_uri)

        payload = {
            'database_uri': database_uri,
            'e_release': e_release,
            'eg_release': eg_release,
            'release_date': release_date,
            'current_release': current_release,
            'email': email,
            'comment': comment,
            'source': source
        }
        return super(MetadataClient, self).submit_job(payload)

    def list_jobs(self, email, job_id, pattern):
        """
        Find jobs and print results
        Arguments:
            email (optional) - filter for the email field
            pattern (optional) - pattern to match against the comment field
        """
        jobs = super(MetadataClient, self).list_jobs()
        filtered = []

        # Filtering is additive; do email first...
        if email is not None:
            for job in jobs:
                if job['input']['email'] == email:
                    filtered.append(job)
        else:
            filtered = jobs

        # ... then look for IDs matching, or later than, the given job_id...
        if job_id is not None:
            filtered[:] = [job for job in filtered if job['id'] >= int(job_id)]

        # ... and then prune based on the comment pattern.
        if pattern is not None:
            comment_pattern = re.compile(pattern)
            filtered[:] = [job for job in filtered if comment_pattern.search(job['input']['comment'])]

        return filtered

    def print_job(self, job, print_results=False, print_input=False):
        logging.info("Job %s (%s) to (%s) - %s" % (
            job['id'], job['input']['metadata_uri'], job['input']['database_uri'], job['status']))
        if print_input:
            self.print_inputs(job['input'])
        if job['status'] == 'complete':
            if print_results:
                logging.info("Load result: " + str(job['status']))
        elif job['status'] == 'running':
            if print_results:
                logging.info("Load result: " + str(job['status']))
                logging.info(str(job['progress']['complete']) + "/" + str(job['progress']['total']) + " task complete")
                logging.info("Status: " + str(job['progress']['message']))
        elif job['status'] == 'failed':
            failure_msg = self.retrieve_job_failure(job['id'])
            logging.info("Job failed with error: " + str(failure_msg['msg']))

    def print_inputs(self, i):
        logging.info("Database URI: " + i['database_uri'])
        if 'e_release' in i:
            logging.info("Ensembl release number: " + str(i['e_release']))
        if 'eg_release' in i:
            logging.info("EG/RR release number: " + str(i['eg_release']))
        if 'release_date' in i:
            logging.info("Release date: " + i['release_date'])
        if 'current_release' in i:
            logging.info("Current release: " + str(i['current_release']))
        logging.info("Email of submitter: " + i['email'])
        logging.info("Comment: " + i['comment'])
        logging.info("Source: " + i['source'])
        logging.info("Submitted: " + i['timestamp'])



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.import logging

import logging
from logging.handlers import SMTPHandler, TimedRotatingFileHandler

log_format_default = '[%(asctime)s] %(levelname)s %(module)s: %(funcName)s(%(lineno)d): %(message)s'

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    datefmt='%m-%d %H:%M:%S')


def get_app_mail_logger(app_name):
    # TODO find SMTP configuration
    mail_handler = SMTPHandler(
        mailhost='127.0.0.1',
        fromaddr='server-error@example.com',
        toaddrs=['admin@example.com'],
        subject='Application Error'
    )
    mail_handler.setLevel(logging.ERROR)
    mail_handler.setFormatter(logging.Formatter(
        '[%(asctime)s] %(levelname)s in %(module)s: %(message)s'
    ))


def file_handler(app_name):
    rotating_file_handler = TimedRotatingFileHandler(
        filename='logs/%s.log' % app_name,
        when='midnight',
        backupCount=5
    )
    rotating_file_handler.setLevel(logging.WARNING)
    rotating_file_handler.setFormatter(
        logging.Formatter(log_format_default))
    return rotating_file_handler


def default_handler():
    stream = logging.StreamHandler()
    stream.setLevel(logging.INFO)
    return stream



#    See the NOTICE file distributed with this work for additional information
#    regarding copyright ownership.
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#        http://www.apache.org/licenses/LICENSE-2.0
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import requests
from ensembl.production.core.server_utils import assert_http_uri
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


class RestClient(object):
    """
    Base client for interacting with a standard production REST service where the URIs meet a common standard.
    Most methods are stubs for overriding or decoration by classes that extend this for specific services
    """

    jobs = '{}jobs'
    jobs_id = '{}jobs/{}'


    def __init__(self, uri):
        assert_http_uri(uri)
        self.uri = uri
        self._http_adapter = self._make_HTTPAdapter()

    def _make_HTTPAdapter(self):
        retries = Retry(total=3, backoff_factor=1,
                        status_forcelist=[429, 500, 502, 503, 504],
                        allowed_methods=["GET", "PUT", "POST", "DELETE"])
        adapter = HTTPAdapter(max_retries=retries)
        return adapter

    def _session(self, use_ssl=False):
        http = requests.Session()
        if use_ssl:
            http.mount("https://", self._http_adapter)
        else:
            http.mount("http://", self._http_adapter)
        http.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        return http

    def submit_job(self, payload):
        """
        Submit a job using the supplied dict as payload. No checking is carried out on the payload
        Arguments:
          payload : job input as dict
        """
        logging.info("Submitting job")
        logging.debug(payload)
        with self._session() as session:
            r = session.post(self.jobs.format(self.uri), json=payload)
        if r.status_code != 201:
            logging.error("failed to submit because: %s", r.text)
        r.raise_for_status()
        return r.json()['job_id']

    def delete_job(self, job_id, kill=False):
        """
        Delete job
        Arguments:
          job_id - ID of job to kill
          kill - if True, job process should be killed
        """
        delete_uri = self.jobs_id.format(self.uri, str(job_id))
        if kill:
            params = {'kill': '1'}
        else:
            params = {}
        with self._session() as session:
            r = session.delete(delete_uri, params=params)
        if r.status_code != 204:
            logging.error("failed to delete job because: %s", r.text)
        r.raise_for_status()
        return True

    def list_jobs(self):
        """
        Find all current jobs
        """
        logging.info("Listing")
        with self._session() as session:
            r = session.get(self.jobs.format(self.uri))
        if r.status_code != 200:
            logging.error("failed to list jobs because: %s", r.text)
        r.raise_for_status()
        return r.json()

    def retrieve_job_failure(self, job_id):
        """
        Retrieve information on a job using the special format "failure" which renders failures from the supplied job.
        The service will respond if it supports this format.
        Arguments:
          job_id - ID of job to retrieve
        """
        logging.info("Retrieving job failure for job %s", job_id)
        with self._session() as session:
            r = session.get(self.jobs_id.format(self.uri, str(job_id)), params={'format': 'failures'})
        if r.status_code != 200:
            logging.error("failed to retrieve job failures because: %s", r.text)
        r.raise_for_status()
        failure_msg = r.json()
        return failure_msg

    def retrieve_job_email(self, job_id):
        """
        Retrieve information on a job using the special format "email" which renders the supplied job in a format suitable
        for sending by email.
        The service will respond if it supports this format.
        Arguments:
          job_id - ID of job to retrieve
        """
        logging.info("Retrieving job as email for job %s", job_id)
        with self._session() as session:
            r = session.get(self.jobs_id.format(self.uri, str(job_id)), params={'format': 'email'})
        r.raise_for_status()
        return r.json()

    def retrieve_job(self, job_id):
        """
        Retrieve information on a job.
        Arguments:
          job_id - ID of job to retrieve
        """
        logging.info("Retrieving results for job %s", job_id)
        with self._session() as session:
            r = session.get(self.jobs_id.format(self.uri, str(job_id)))
        if r.status_code != 200:
            logging.error("failed to retrieve job because: %s", r.text)
        r.raise_for_status()
        job = r.json()
        return job

    def print_job(self, job, **kwargs):
        """
        Stub utility to print job to logging
        Arguments:
          job - job object
          print_results - ignored
          print_input - ignored
        """
        logging.info(job)

    def write_output(self, r, output_file):
        """
        Utility to write response.
        Arguments:
          job - response object
          output_file - output file handle
        """
        if output_file is not None:
            with output_file as f:
                f.write(r.text)

